/* 
 * Functions.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "Functions.h"
#include "GNUstep.h"

#ifndef max
#define max(a,b) ((a) >= (b) ? (a):(b))
#endif

#ifndef min
#define min(a,b) ((a) <= (b) ? (a):(b))
#endif

extern float *earthRadius;

float radiusForLatitude(float lat)
{
  float l, r, sc, x, y, z, e2;
  
  e2 = ECCENTRICITY * ECCENTRICITY;
  l = degrToRad(lat);
  sc = sin(l);
  x = EQUAT_RADIUS * (1.0 - e2);
  z = 1.0 - e2 * sc * sc;
  y = pow(z, 1.5);
  r = x / y;

  return r * 1000.0;
}

NSPoint screenPointForEarthPoint(NSPoint ep, NSPoint cp, long scale)
{
  PJ *ref;
  UV data;  
  char *parms[PARMSN];
  float x, y;
  int i;

  for (i = 0; i < PARMSN; i++) {
    parms[i] = NSZoneMalloc(NSDefaultMallocZone(), sizeof(char) * PARMSZ);
  }  
  
  sprintf(parms[0], "proj=aeqd");
  sprintf(parms[1], "ellps=WGS84");
  sprintf(parms[2], "lon_0=%f", cp.x);
  sprintf(parms[3], "lat_0=%f", cp.y);

  if (!(ref = pj_init(PARMSN, parms))) {
    return NSZeroPoint;
  } 

  data.u = ep.x * DEG_TO_RAD;
  data.v = ep.y * DEG_TO_RAD;
  
  data = pj_fwd(data, ref);

  if ((data.u == HUGE_VAL) || (data.v == HUGE_VAL)) {
    return NSZeroPoint;
  }  
  
  x = (MAP_W / 2) + data.u / (scale / PIXELFACT); 
  y = (MAP_H / 2) + data.v / (scale / PIXELFACT);
    
  pj_free(ref);
  
  for (i = 0; i < PARMSN; i++) {
    NSZoneFree(NSDefaultMallocZone(), parms[i]);
  }
    
  return NSMakePoint(x, y);
}

NSPoint earthPointForScreenPoint(NSPoint sp, NSPoint cp, long scale)
{
  PJ *ref;
  UV data;  
  char *parms[PARMSN];
  float x, y;
  int i;

  for (i = 0; i < PARMSN; i++) {
    parms[i] = NSZoneMalloc(NSDefaultMallocZone(), sizeof(char) * PARMSZ);
  }  

  sprintf(parms[0], "proj=aeqd");
  sprintf(parms[1], "ellps=WGS84");
  sprintf(parms[2], "lon_0=%f", cp.x);
  sprintf(parms[3], "lat_0=%f", cp.y);

  if (!(ref = pj_init(PARMSN, parms))) {
    return NSZeroPoint;
  } 

  data.u = ((MAP_W / 2) - sp.x) * (scale / PIXELFACT); 
  data.u = cp.x - data.u;
  data.v = ((MAP_H / 2) - sp.y) * (scale / PIXELFACT);  
  data.v = cp.y - data.v;

  data = pj_inv(data, ref);

  x = data.u * RAD_TO_DEG;
  y = data.v * RAD_TO_DEG;

  pj_free(ref);

  for (i = 0; i < PARMSN; i++) {
    NSZoneFree(NSDefaultMallocZone(), parms[i]);
  }

  return NSMakePoint(x, y);
}

BOOL earthPointFits(NSPoint ep, NSPoint cp, long scale)
{
  NSPoint sp = screenPointForEarthPoint(ep, cp, scale);
  return (sp.x > BORDER && sp.x < MAP_W - BORDER 
                    && sp.y > BORDER && sp.y < MAP_H - BORDER);
}

NSString *coveredAreaForScale(long scale, int units)
{
  float spacex, spacey, fact;
  NSString *unitStr;
  
  if (units == METRIC) {
    fact = 1;
    unitStr = @"km";
  } else if (units == MILES) {
    fact = KM2MILES;
    unitStr = @"miles";  
  } else {
    fact = KM2NAUTIC;
    unitStr = @"nautic";    
  }
  
  spacex = scale * 0.640 / PIXELFACT * fact;
  spacey = scale * 0.512 / PIXELFACT * fact;
  
  return [NSString stringWithFormat: @"%.3f x %.3f %@", spacex, spacey, unitStr];
}

float distanceBetweenPoints(NSPoint actp, NSPoint dstp)
{
  return (((earthRadius[(int)(100 + actp.y)] 
            + earthRadius[(int)(100 + dstp.y)]) / 2)
             * acos(sindegr(actp.y) * sindegr(dstp.y) 
                  + cosdegr(actp.y) * cosdegr(dstp.y) * 
                    cos(degrToRad(dstp.x) - degrToRad(actp.x)))) / 1000;
}

float angleBetweenPoints(NSPoint actp, NSPoint dstp)
{
  float dist = distanceBetweenPoints(actp, dstp);
  NSPoint p = NSMakePoint(dstp.x, actp.y);
  float cat = distanceBetweenPoints(p, dstp);
  float ang = asin(cat / dist) / PI * 180;
    
  if (actp.x >= dstp.x) {
    if (actp.y >= dstp.y) {
      ang = 90 - ang;
    } else {
      ang += 90;
    }
  } else {
    if (actp.y >= dstp.y) {
      ang += 270;
    } else {
      ang = 270 - ang;
    }
  }

  return ang;
}

NSMenuItem *addItemToMenu(NSMenu *menu, NSString *str, 
															NSString *comm, NSString *sel, NSString *key)
{
	NSMenuItem *item = [menu addItemWithTitle: NSLocalizedString(str, comm)
												action: NSSelectorFromString(sel) keyEquivalent: key]; 
	return item;
}
