/* 
 * GPSInterface.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <AppKit/AppKit.h>
#include "GPSInterface.h"
#include "Position.h"
#include "GNUstep.h"
#include <math.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

#define TIMEOUT 10
#define NMEA_LENGTH 80

@implementation NSFileHandle (NSFileHandleSocketConnection)

+ (id)fileHandleForConnectionToHost:(NSString *)hostname 
                             atPort:(unsigned)port
{
  struct hostent *hostinfo;
  struct sockaddr_in remoteAddr;
  int sock;
  int i;

  // Resolve the hostname
  hostinfo = gethostbyname([hostname cString]);
  if (hostinfo == NULL) {
    return nil;
  }

  // Retrieve a socket
  sock = socket(PF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    return nil;
  }
  
  // Try to make a connection to the host by trying all possible addresses
  bzero((char*)&remoteAddr, sizeof(remoteAddr));
  remoteAddr.sin_family = AF_INET;
  remoteAddr.sin_port = htons(port);
  
  for (i = 0; hostinfo->h_addr_list[i] != NULL; i++) {
    remoteAddr.sin_addr = *(struct in_addr*)hostinfo->h_addr_list[i];
    
    if (connect(sock, (struct sockaddr*)&remoteAddr, sizeof(remoteAddr)) == 0) {
      // Connection was succesfull
      return AUTORELEASE ([[NSFileHandle alloc] initWithFileDescriptor: sock
                                                        closeOnDealloc: YES]);
    }
  }

  return nil;
}

@end


@implementation GPSInterface

- (void)dealloc
{
  if (timeoutTimer && [timeoutTimer isValid]) {
    [timeoutTimer invalidate];
  }
  
  [nc removeObserver: self];
  
  TEST_RELEASE (socketHandle);
  TEST_RELEASE (timeoutDate);  
  RELEASE (nmeaSentence);
  
  RELEASE (satellites);
  RELEASE (date);
  RELEASE (lastUpdate);
  
  [super dealloc];	
}

- (id)init
{  
	self = [super init];

  if (self) {
    pos = [Position position];
    nc = [NSNotificationCenter defaultCenter];
    socketHandle = nil;
    timeoutDate = nil;
    timeoutTimer = nil;
    nmeaSentence = [NSMutableString new];
    
    ASSIGN (date, [NSCalendarDate date]);
    ASSIGN (lastUpdate, [NSDate date]);
    
    satellites = [NSMutableArray new];
    visibleSats = 0;
    usedSats = 0;  
    status = 0;
    satsDone = NO;
    mode = 0;
    speed = 0;
    track = 0;
    latitude = 0;
    longitude = 0;
    altitude = 0;
    pdop = 0;
    hdop = 0;
    vdop = 0;  
  }
    
	return self;
}

- (BOOL)openGpsdConnectionToHost:(NSString *)hostname 
                          atPort:(NSString *)portname
                        gpsdPath:(NSString *)gpsdpath
                        gpsdArgs:(NSString *)gpsdargs
{
  NSHost *host = [NSHost hostWithName: hostname];

  [self closeConnection];

  if ([host isEqual: [NSHost currentHost]] 
                          || [hostname isEqual: @"localhost"]) {
    socketHandle = [NSFileHandle fileHandleForConnectionToHost: hostname 
                                                        atPort: [portname intValue]];

    if (socketHandle == nil) {
      NSArray *args = [gpsdargs componentsSeparatedByString: @" "];
      NSDate *when = nil;
      BOOL done = NO;
    
      [NSTask launchedTaskWithLaunchPath: gpsdpath arguments: args];
    
      while (done == NO) {
	      if (when == nil) {
		      when = [[NSDate alloc] init];
		    } else if ([when timeIntervalSinceNow] > 5) {
		      DESTROY (when);
          done = YES;
		    }

        if (socketHandle == nil && done == NO) {
		      NSDate *limit = [[NSDate alloc] initWithTimeIntervalSinceNow: 0.5];
		      [[NSRunLoop currentRunLoop] runUntilDate: limit];
		      RELEASE(limit);
                    
          socketHandle = [NSFileHandle fileHandleForConnectionToHost: hostname 
                                                              atPort: [portname intValue]];
        } else {
          done = YES;
        }
      }
      
      TEST_RELEASE (when);
    }

  } else {
    socketHandle = [NSFileHandle fileHandleForConnectionToHost: hostname 
                                                        atPort: [portname intValue]];
  }

  if (socketHandle) {
    RETAIN (socketHandle);
    
    [socketHandle writeData: [NSData dataWithBytes: "R\n" length: 2]];

	  [nc addObserver: self 
      	   selector: @selector(socketData:) 
      			   name: NSFileHandleReadCompletionNotification
      		   object: (id)socketHandle];

	  [socketHandle readInBackgroundAndNotify];
    
    timeoutDate = [[NSDate alloc] initWithTimeIntervalSinceNow: TIMEOUT];

    timeoutTimer = [NSTimer scheduledTimerWithTimeInterval: 5.0 
												 target: self selector: @selector(checkConnection:) 
																								  userInfo: nil repeats: YES];
    return YES;    
  }
  
  return NO;
}

- (void)checkConnection:(id)sender
{
  NSDate *now = [NSDate date];
    
  if ([now laterDate: timeoutDate] == now) {
    [self closeConnection];
    [pos gpsdConnectionLost];
  }
}

- (void)closeConnection
{
  if (socketHandle) {
    [socketHandle closeFile];
    DESTROY (socketHandle);
  }

  DESTROY (timeoutDate);
  
  if (timeoutTimer && [timeoutTimer isValid]) {
    [timeoutTimer invalidate];
  }
}

- (void)socketData:(NSNotification *)notif
{
  NSDictionary *userInfo = [notif userInfo];
  NSData *data = [userInfo objectForKey: NSFileHandleNotificationDataItem];
  int dlength = [data length];
  int i;

  DESTROY (timeoutDate);
  timeoutDate = [[NSDate alloc] initWithTimeIntervalSinceNow: TIMEOUT];
  
  if (dlength) {
    NSString *token = [[NSString alloc] initWithData: data 
																					  encoding: NSASCIIStringEncoding];
    
    for (i = 0; i < dlength; i++) {
      char c = [token characterAtIndex: i];

      if (c == '$') {
        if (nmeaSentence) {
          NSString *checked = [self checkSentence];

          if (checked) {
            NSString *type = [self processSentence: checked];
            
            if ([type isEqual: @"GPRMC"]) {
              [pos setLongitude: longitude latitude: latitude];
              [pos setSpeed: speed];
              [pos setTrack: track];
              [pos setGpsDate: date];

            } else if ([type isEqual: @"GPGGA"]) {
              [pos setLongitude: longitude latitude: latitude];
              [pos setAltitude: altitude];
              [pos setUsedSats: usedSats];
              
              // NSLog(@"status = %i", status);
              // NSLog(@"usedSats = %i", usedSats);
             
            } else if ([type isEqual: @"GPGSA"]) {
              // NSLog(@"mode = %i", mode);
              // NSLog(@"pdop = %f", pdop);
              // NSLog(@"hdop = %f", hdop);
              // NSLog(@"vdop = %f", vdop);
            
            } else if ([type isEqual: @"GPGSV"] && satsDone) {
              [pos updateSatellites: satellites];
            }
          }
          
          DESTROY (nmeaSentence);
          nmeaSentence = [NSMutableString new];
        }        
      }
      
      if (nmeaSentence) {
        if ((c != '$') && (c != '\n') && c != '\r') {
          [nmeaSentence appendFormat: @"%c", c];
        }
      }      
    }
    
    RELEASE (token);
  }  
    
  [socketHandle readInBackgroundAndNotify];
}

- (NSString *)checkSentence
{
  NSMutableString *checked = [NSMutableString string];
  const char *sentence = [nmeaSentence cString];
  char *p = (char *)sentence;
  unsigned char sum = '\0';
  char csum[3];
  char c;

  while ((c = *p++) != '*') {
	  sum ^= c;
    [checked appendFormat: @"%c", c];
  }
  
  sprintf(csum, "%02X", sum);
  
  if (strncmp(csum, p, 2) == 0) {
    return checked;
  }

  return nil;
}

- (NSString *)processSentence:(NSString *)sentence
{
  NSArray *parts = [sentence componentsSeparatedByString: @","];
  NSString *sname = [parts objectAtIndex: 0];
  
  if ([sname isEqual: @"GPRMC"]) {
    [self processGPRMC: parts];
  } else if ([sname isEqual: @"GPGGA"]) {
    [self processGPGGA: parts];
  } else if ([sname isEqual: @"GPGSA"]) {
    [self processGPGSA: parts];
  } else if ([sname isEqual: @"GPGSV"]) {
    [self processGPGSV: parts];
  } 
  
  return sname;
}

- (void)processGPRMC:(NSArray *)parts
{
  int updated = 0;
  NSString *s;
  NSString *ts;
  float oldlat, oldlon;
  double m, d;
      
  s = [parts objectAtIndex: 9];
  ts = [parts objectAtIndex: 1];
  s = [s stringByAppendingString: ts];
  ASSIGN (date, [NSCalendarDate dateWithString: s 
                                calendarFormat: @"%d%m%y%H%M%S"]);

  s = [parts objectAtIndex: 2];
  if ([s isEqual: @"V"]) {
    status = 0;
  }

  s = [parts objectAtIndex: 7];
  speed = [s floatValue];
  
  s = [parts objectAtIndex: 8];
  track = [s floatValue];
  
  s = [parts objectAtIndex: 3];
  oldlat = latitude;
  latitude = [s floatValue];
  m = 100.0 * modf(latitude / 100.0, &d);
  latitude = d + m / 60.0;
  
  s = [parts objectAtIndex: 4];
  if ([s isEqual: @"S"]) {
    latitude = -latitude;
  }
  
  if (latitude != oldlat) {
    updated++;
  } 

  s = [parts objectAtIndex: 5];
  oldlon = longitude;
  longitude = [s floatValue];
	m = 100.0 * modf(longitude / 100.0, &d);
	longitude = d + m / 60.0;
  
  s = [parts objectAtIndex: 6];
  if ([s isEqual: @"W"]) {
    longitude = -longitude;
  }

  if (longitude != oldlon) {
    updated++;
  } 
    
  if (updated == 2) {
    ASSIGN (lastUpdate, [NSDate date]);
  }
}

- (void)processGPGGA:(NSArray *)parts
{
  int updated = 0;
  NSString *s;
  float oldlat, oldlon;
  double m, d;

  s = [parts objectAtIndex: 2];
  oldlat = latitude;
  latitude = [s floatValue];
  m = 100.0 * modf(latitude / 100.0, &d);
  latitude = d + m / 60.0;
  
  s = [parts objectAtIndex: 3];
  if ([s isEqual: @"S"]) {
    latitude = -latitude;
  }
  
  if (latitude != oldlat) {
    updated++;
  } 

  s = [parts objectAtIndex: 4];
  oldlon = longitude;
  longitude = [s floatValue];
	m = 100.0 * modf(longitude / 100.0, &d);
	longitude = d + m / 60.0;
  
  s = [parts objectAtIndex: 5];
  if ([s isEqual: @"W"]) {
    longitude = -longitude;
  }

  if (longitude != oldlon) {
    updated++;
  } 
    
  if (updated == 2) {
    ASSIGN (lastUpdate, [NSDate date]);
  }

  s = [parts objectAtIndex: 6];
  status = [s intValue];

  s = [parts objectAtIndex: 7];
  usedSats = [s intValue];

  s = [parts objectAtIndex: 9];
  altitude = [s floatValue];
}

- (void)processGPGSA:(NSArray *)parts
{
  NSString *s;

  s = [parts objectAtIndex: 2];
  mode = [s intValue];

  s = [parts objectAtIndex: 15];
  pdop = [s floatValue];

  s = [parts objectAtIndex: 16];
  hdop = [s floatValue];

  s = [parts objectAtIndex: 17];
  vdop = [s floatValue];
}

- (void)processGPGSV:(NSArray *)parts
{
  NSString *s;
  int messages;
  int messIndex;
  int satind = 4;
  int last = [parts count] - 4;
  
  s = [parts objectAtIndex: 3];
  visibleSats = [s intValue];

  s = [parts objectAtIndex: 1];
  messages = [s intValue];

  s = [parts objectAtIndex: 2];
  messIndex = [s intValue];

  if (messIndex == 1) {
    [satellites removeAllObjects];
    satsDone = NO;
  } else if (messIndex == messages) {
    satsDone = YES;
  }

  while (satind <= last) {
    NSString *sname = [parts objectAtIndex: satind];
    NSMutableDictionary *dict = [self satWithName: sname];
    
    if (dict == nil) {
      dict = [self addSatWithName: sname];
    }
        
    s = [parts objectAtIndex: satind + 1];
    if ([s length]) {
      [dict setObject: [NSNumber numberWithInt: [s intValue]] 
                                        forKey: @"elevation"];
    }
    s = [parts objectAtIndex: satind + 2];
    if ([s length]) {
      [dict setObject: [NSNumber numberWithInt: [s intValue]] 
                                        forKey: @"azimuth"];
    }
    s = [parts objectAtIndex: satind + 3];
    if ([s length]) {
      [dict setObject: [NSNumber numberWithInt: [s intValue]] 
                                        forKey: @"strength"];
    }
    
    satind += 4;
  }
}

- (NSMutableDictionary *)satWithName:(NSString *)name
{
  int i;
  
  for (i = 0; i < [satellites count]; i++) {
    NSMutableDictionary *dict = [satellites objectAtIndex: i];
    NSString *satName = [dict objectForKey: @"name"];
  
    if ([satName isEqual: name]) {
      return dict;
    }
  }

  return nil;
}

- (NSMutableDictionary *)addSatWithName:(NSString *)name
{
  NSMutableDictionary *dict = [NSMutableDictionary dictionary];

  [dict setObject: name forKey: @"name"];
  [dict setObject: [NSNumber numberWithInt: 0] forKey: @"elevation"];
  [dict setObject: [NSNumber numberWithInt: 0] forKey: @"azimuth"];
  [dict setObject: [NSNumber numberWithInt: 0] forKey: @"strength"];

  [satellites insertObject: dict atIndex: [satellites count]];
  
  return dict;
}

- (void)getLatitude:(float *)lat longitude:(float *)lon
{
  *lat = latitude;
  *lon = longitude;
}

@end
