/* 
 * InspectorsController.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include "InspectorsController.h"
#include "InspectorsProtocol.h"
#include "SatsInspector.h"
#include "NavigationInspector.h"
#include "WayPointInspector.h"
#include "WayPoint.h"
#include "RouteInspector.h"
#include "Position.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"

#ifdef GNUSTEP 
  #define WINHEIGHT 446
#else
  #define WINHEIGHT 428
#endif

static NSString *nibName = @"InspectorsWin";

@implementation InspectorsController

- (void)dealloc
{
	TEST_RELEASE (win);
	TEST_RELEASE (inspectors);
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: nibName owner: self] == NO) {
      NSLog(@"Inspectors Controller: failed to load %@!", nibName);
    } else {  
      int index;
      
      if ([win setFrameUsingName: @"inspectorsWin"] == NO) {
        [win setFrame: NSMakeRect(100, 100, 276, WINHEIGHT) display: NO];
      }
      [win setDelegate: self];  
      [inspTabView setDelegate: self];  

		  inspectors = [[NSMutableArray alloc] initWithCapacity: 1];
      index = 0;
      
      currentInspector = (id<InspectorsProtocol>)[[SatsInspector alloc] init];	
      [inspectors insertObject: currentInspector atIndex: index];
      [[inspTabView tabViewItemAtIndex: index] setView: [currentInspector inspView]];
      DESTROY (currentInspector);
      index++;
      
      currentInspector = (id<InspectorsProtocol>)[[NavigationInspector alloc] init];	
      [inspectors insertObject: currentInspector atIndex: index]; 
      [[inspTabView tabViewItemAtIndex: index] setView: [currentInspector inspView]];
      DESTROY (currentInspector);
      index++;
      
      currentInspector = (id<InspectorsProtocol>)[[WayPointInspector alloc] init];	
      [inspectors insertObject: currentInspector atIndex: index]; 
      [[inspTabView tabViewItemAtIndex: index] setView: [currentInspector inspView]];
      DESTROY (currentInspector);
      index++;
      
      currentInspector = (id<InspectorsProtocol>)[[RouteInspector alloc] init];	
      [inspectors insertObject: currentInspector atIndex: index]; 
      [[inspTabView tabViewItemAtIndex: index] setView: [currentInspector inspView]];
      DESTROY (currentInspector);

      /* Internationalization */
      [lonLabel setStringValue: NSLocalizedString(@"longitude", @"")];
      [latLabel setStringValue: NSLocalizedString(@"latitude", @"")];
      [altLabel setStringValue: NSLocalizedString(@"altitude", @"")];
      [unitsLabel setStringValue: NSLocalizedString(@"units", @"")];
      [distLabel setStringValue: NSLocalizedString(@"dist. to target", @"")];
      [speedLabel setStringValue: NSLocalizedString(@"speed", @"")];
      [destLabel setStringValue: NSLocalizedString(@"time at dest.", @"")];
      [timeLabel setStringValue: NSLocalizedString(@"gps time", @"")];

      [popUp setEnabled: YES];
      [self activateInspector: popUp];
    } 
  }
  
  return self;
}

- (IBAction)activateInspector:(id)sender
{
  int index = [sender indexOfSelectedItem];
  
  if ([inspectors count] <= index) {
    return;
  }
	if (currentInspector 
        && (currentInspector == [inspectors objectAtIndex: index])) {
    return;
	}

  currentInspector = [inspectors objectAtIndex: index];
	[win setTitle: [currentInspector winname]];
  
  [inspTabView selectTabViewItemAtIndex: index];
      
  if ([currentInspector isKindOfClass: [WayPointInspector class]]) {
    [[NSNotificationCenter defaultCenter] 
        postNotificationName: GPSSelectedWayPointDidChange 
                      object: nil];
  }
}

- (void)setGpsTime:(NSString *)timeStr
{
  [timeField setStringValue: timeStr];
}

- (void)setLongitude:(float)lon latitude:(float)lat
{
  [lonField setStringValue: [NSString stringWithFormat: @"%f", lon]];
  [latField setStringValue: [NSString stringWithFormat: @"%f", lat]];
}

- (void)setAltitude:(float)alt
{
  [altField setStringValue: [NSString stringWithFormat: @"%.2f", alt]];
}

- (void)setSpeed:(float)speed
{
  [speedField setStringValue: [NSString stringWithFormat: @"%.2f", speed]];
}

- (void)setDistToTarget:(float)dist
{
  [distField setStringValue: [NSString stringWithFormat: @"%.3f", dist]];
}

- (void)showSatellites
{
  [popUp selectItemAtIndex: 0];
  [self activateInspector: popUp];
}

- (void)showNavigation
{
  [popUp selectItemAtIndex: 1];
  [self activateInspector: popUp];
}

- (void)showWayPoints
{
  [popUp selectItemAtIndex: 2];
  [self activateInspector: popUp];
}

- (void)showRoutes
{
  [popUp selectItemAtIndex: 3];
  [self activateInspector: popUp];
}

- (void)updateDefaults
{
  [win saveFrameUsingName: @"inspectorsWin"];
}

- (id)myWin
{
  return win;
}

- (BOOL)windowShouldClose:(id)sender
{
  [self updateDefaults];
	return YES;
}

//
// NSTabView delegate Methods
//
- (BOOL)tabView:(NSTabView *)tabView shouldSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
  return YES;
}

- (void)tabView:(NSTabView *)tabView willSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
}

- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
}

- (void)tabViewDidChangeNumberOfTabViewItems:(NSTabView *)TabView
{
}
	
@end
