/* 
 * NavigationInspector.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "NavigationInspector.h"
#include "Position.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"

#ifndef max
#define max(a,b) ((a) >= (b) ? (a):(b))
#endif

#ifndef min
#define min(a,b) ((a) <= (b) ? (a):(b))
#endif

#define IM_RADIUS 74.0
#define LLENGHT 20.0
#define LINEOR 54.0
#define CHAROR 62.0
#define CHARSHFTX 6.0
#define CHARSHFTY 8.0
#define ARRH 80.0
#define ARRW 30.0
#define REDARRH (ARRH - 8)
#define REDARRW (ARRW - 12)

static NSString *inspectorNibName = @"NavigationInspector";

@implementation NavigationInspector

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  TEST_RELEASE (navView);
  TEST_RELEASE (mainBox);
  
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: inspectorNibName owner: self] == NO) {
      NSLog(@"failed to load %@!", inspectorNibName);
    } else {     
      RETAIN (mainBox);
      RELEASE (win);
      
      navView = [[NavigationView alloc] init];
      [(NSBox *)topBox setContentView: navView];
      
      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(setTrack:) 
                					  name: GPSSetTrackNotification
                					object: nil];

      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(setGotoDir:) 
                					  name: GPSSetGotoDirNotification
                					object: nil];
    }
  }
  
  return self;
}

- (void)setTrack:(NSNotification *)notif
{
  [navView setDirection: [[notif object] floatValue]];
}

- (void)setGotoDir:(NSNotification *)notif
{
  [navView setGotoDirection: [[notif object] floatValue]];
}

- (id)inspView
{
  return mainBox;
}

- (void)deactivate
{
  [mainBox removeFromSuperview];
}

- (NSString *)inspname
{
	return NSLocalizedString(@"Navigation", @"");
}

- (NSString *)winname
{
	return NSLocalizedString(@"Navigation Inspector", @"");
}
	
@end

@implementation NavigationView

- (void)dealloc
{
  RELEASE (image);
  RELEASE (font);
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
    ASSIGN (image, [NSImage imageNamed: @"navigationCircle"]);
    #ifdef GNUSTEP    
      ASSIGN (font, [NSFont userFixedPitchFontOfSize: 24]);
    #else
      ASSIGN (font, [NSFont userFixedPitchFontOfSize: 20]);
    #endif
    direction = 0.0;
    gotoDirect = 0.0;
    impoint = NSZeroPoint;
    center = NSZeroPoint;
    pos = [Position position];
  }
  
  return self;
}

- (void)setDirection:(float)dir
{
  if ((max(direction, dir) - min(direction, dir)) > 1) {
    direction = dir;
    [self setNeedsDisplay: YES];
  }
}

- (void)setGotoDirection:(float)dir
{
  if ((max(gotoDirect, dir) - min(gotoDirect, dir)) > 1) {
    gotoDirect = dir;
    [self setNeedsDisplay: YES];
  }
}

- (void)setFrame:(NSRect)frameRect
{
  NSSize ssize = frameRect.size;
  NSSize isize = [image size];
   
  center = NSMakePoint(ssize.width / 2, ssize.height / 2);
  impoint = NSMakePoint(ssize.width / 2 - isize.width / 2, 
                                  ssize.height / 2 - isize.height / 2);

  [super setFrame: frameRect];
}

- (void)drawRect:(NSRect)rect
{
#ifdef GNUSTEP     
  NSGraphicsContext *context = [NSGraphicsContext currentContext];
#endif
  const char *cardpts[4] = { "N", "W", "S", "E" };
  NSBezierPath *bpath;
  NSAffineTransform *t1, *t2, *t3;
  NSPoint p[2];
  float f;
  int i;
    
  [image compositeToPoint: impoint operation: NSCompositeSourceOver];

  f = 360 - direction;
  
  [font set];
  
  for (i = 0; i < 4; i++) {
    p[0].x = CHAROR * sindegr(f);
    p[0].x += center.x;
    p[0].x -= CHARSHFTX;
    p[0].y = CHAROR * cosdegr(f);
    p[0].y += center.y;
    p[0].y -= CHARSHFTY;

  #ifdef GNUSTEP     
    DPSmoveto(context, p[0].x, p[0].y);
    DPSshow(context, cardpts[i]);
  #else
    {
      NSString *str = [NSString stringWithCString: cardpts[i]];    
      bpath = [NSBezierPath bezierPath]; 
      [bpath moveToPoint: p[0]];
      [bpath appendBezierPathWithGlyph: [font glyphWithName: str] inFont: font];
      [bpath stroke];
      [bpath fill];
    }
  #endif
      
    f -= 45;
  
    p[0].x = LINEOR * sindegr(f);
    p[0].x += center.x;
    p[0].y = LINEOR * cosdegr(f);
    p[0].y += center.y;
    
    p[1].x = IM_RADIUS * sindegr(f);
    p[1].x += center.x;
    p[1].y = IM_RADIUS * cosdegr(f);
    p[1].y += center.y;
  
    [NSBezierPath strokeLineFromPoint: p[0] toPoint: p[1]];
  
    f -= 45;
  }
  
  bpath = [NSBezierPath bezierPath]; 
  [bpath moveToPoint: center];
  
  [bpath relativeLineToPoint: NSMakePoint(0, -(ARRH / 2))];
  [bpath relativeLineToPoint: NSMakePoint(-(ARRW / 4), 0)];
  [bpath relativeLineToPoint: NSMakePoint(0, (ARRH / 2))];
  [bpath relativeLineToPoint: NSMakePoint(-(ARRW / 4), 0)];
  [bpath relativeLineToPoint: NSMakePoint((ARRW / 2), (ARRH / 2))];
  [bpath relativeLineToPoint: NSMakePoint((ARRW / 2), -(ARRH / 2))];
  [bpath relativeLineToPoint: NSMakePoint(-(ARRW / 4), 0)];
  [bpath relativeLineToPoint: NSMakePoint(0, -(ARRH / 2))];
  [bpath relativeLineToPoint: NSMakePoint(-(ARRW / 4), 0)];
  [bpath relativeLineToPoint: NSMakePoint(0, (ARRH / 2))];
  [bpath closePath];

  [bpath stroke];
  [bpath fill];

  if ([pos followRoute]) {
    bpath = [NSBezierPath bezierPath]; 
    [bpath moveToPoint: center];

    [bpath relativeLineToPoint: NSMakePoint(0, -(REDARRH / 2))];
    [bpath relativeLineToPoint: NSMakePoint(-(REDARRW / 4), 0)];
    [bpath relativeLineToPoint: NSMakePoint(0, (REDARRH / 2))];
    [bpath relativeLineToPoint: NSMakePoint(-(REDARRW / 4), 0)];
    [bpath relativeLineToPoint: NSMakePoint((REDARRW / 2), (REDARRH / 2))];
    [bpath relativeLineToPoint: NSMakePoint((REDARRW / 2), -(REDARRH / 2))];
    [bpath relativeLineToPoint: NSMakePoint(-(REDARRW / 4), 0)];
    [bpath relativeLineToPoint: NSMakePoint(0, -(REDARRH / 2))];
    [bpath relativeLineToPoint: NSMakePoint(-(REDARRW / 4), 0)];
    [bpath relativeLineToPoint: NSMakePoint(0, (REDARRH / 2))];
    [bpath closePath];

    f = 360 - gotoDirect + 180;
      
    t1 = [NSAffineTransform transform];	
    [t1 translateXBy: -center.x yBy: -center.y];
    [bpath transformUsingAffineTransform: t1];

	  t2 = [NSAffineTransform transform];	
    [t2 rotateByDegrees: f];
    [bpath transformUsingAffineTransform: t2];

	  t3 = [NSAffineTransform transform];	
    [t3 translateXBy: center.x yBy: center.y];
    [bpath transformUsingAffineTransform: t3];

    [[NSColor redColor] set];

    [bpath stroke];
    [bpath fill];  
  }
}

@end

