/* 
 * RouteInspector.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "RouteInspector.h"
#include "WayPoint.h"
#include "Position.h"
#include "Notifications.h"
#include "Functions.h"
#include "GNUstep.h"

#define CHECKSIZE(sz) \
if (sz.width < 0) sz.width = 0; \
if (sz.height < 0) sz.height = 0

static NSString *inspectorNibName = @"RouteInspector";

@implementation RouteInspector

- (void)dealloc
{
  TEST_RELEASE (win);
  TEST_RELEASE (routeMatrix);
  TEST_RELEASE (pointsMatrix);
  
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: inspectorNibName owner: self] == NO) {
      NSLog(@"failed to load %@!", inspectorNibName);
    } else {  
      NSArray *routesNames;  
      NSSize cs, ms;
      int i;
      
      pos = [Position position];
      
      [routeScroll setBorderType: NSBezelBorder];
		  [routeScroll setHasHorizontalScroller: NO];
  	  [routeScroll setHasVerticalScroller: YES]; 

      routeMatrix = [[NSMatrix alloc] initWithFrame: NSMakeRect(0, 0, 100, 100)
				            	mode: NSRadioModeMatrix prototype: [NSBrowserCell new]
			       											          numberOfRows: 0 numberOfColumns: 0];
      [routeMatrix setTarget: self];
      [routeMatrix setAction: @selector(routeMatrixAction:)];
      [routeMatrix setIntercellSpacing: NSZeroSize];
      [routeMatrix setCellSize: NSMakeSize(130, 16)];
      [routeMatrix setAutoscroll: YES];
	    [routeMatrix setAllowsEmptySelection: NO];
      cs = [routeScroll contentSize];
      ms = [routeMatrix cellSize];
      ms.width = cs.width;
      CHECKSIZE (ms);
      [routeMatrix setCellSize: ms];
	    [routeScroll setDocumentView: routeMatrix];	

      [pointsScroll setBorderType: NSBezelBorder];
		  [pointsScroll setHasHorizontalScroller: NO];
  	  [pointsScroll setHasVerticalScroller: YES]; 

      routesNames = [[pos routesDict] allKeys];

      if ([routesNames count]) {
        [routeMatrix addColumn];
      }
      
      for (i = 0; i < [routesNames count]; i++) {
        NSString *rname = [routesNames objectAtIndex: i];
        id cell;

        if (i != 0) {
		      [routeMatrix insertRow: i];
        }     
        cell = [routeMatrix cellAtRow: i column: 0];   
        [cell setStringValue: rname];
        [cell setLeaf: YES];
      }
      
      if ([routesNames count]) {
        [routeMatrix sizeToCells]; 
      }
      
      pointsMatrix = [[NSMatrix alloc] initWithFrame: NSMakeRect(0, 0, 100, 100)
				            	mode: NSRadioModeMatrix prototype: [NSBrowserCell new]
			       											          numberOfRows: 0 numberOfColumns: 0];
      [pointsMatrix setIntercellSpacing: NSZeroSize];
      [pointsMatrix setCellSize: NSMakeSize(130, 16)];
      [pointsMatrix setAutoscroll: YES];
	    [pointsMatrix setAllowsEmptySelection: YES];
      cs = [pointsScroll contentSize];
      ms = [pointsMatrix cellSize];
      ms.width = cs.width;
      CHECKSIZE (ms);
      [pointsMatrix setCellSize: ms];
	    [pointsScroll setDocumentView: pointsMatrix];	
      
      [follRouteSwtch setState: NSOffState];
      [follRouteSwtch setEnabled: NO];

      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(endFollowRoute:) 
                					  name: GPSSetGotoDirNotification
                					object: nil];
      
      /* Internationalization */
      [routeLabel setStringValue: NSLocalizedString(@"Routes", @"")];
      [pointsLabel setStringValue: NSLocalizedString(@"Way points", @"")];
      [delRouteButt setTitle: NSLocalizedString(@"delete", @"")];
      [newRouteButt setTitle: NSLocalizedString(@"new", @"")];
      [revRouteButt setTitle: NSLocalizedString(@"reverse", @"")];
      [follRouteSwtch setTitle: NSLocalizedString(@"follow", @"")];
      [delPointButt setTitle: NSLocalizedString(@"remove", @"")];
      [newPointButt setTitle: NSLocalizedString(@"add", @"")];
    }
  }
  
  return self;
}

- (IBAction)newRouteAction:(id)sender
{
  NSString *rname = [routeNameField stringValue];
  NSArray *cells = [routeMatrix cells];
  int count = [cells count];
  id cell;
  int i;
  
  if ([rname length] == 0) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
          NSLocalizedString(@"You must supply a name for the new route!", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
    return;
  }
  
  for (i = 0; i < count; i++) {
    if ([[[cells objectAtIndex: i] stringValue] isEqual: rname]) {
      NSRunAlertPanel(NSLocalizedString(@"error", @""), 
            NSLocalizedString(@"A route with this name already exists!", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
      [routeNameField setStringValue: @""];                                                                  
      return;
    }
  }
  
  if (count == 0) {
    [routeMatrix addColumn];
  } else {
    [routeMatrix insertRow: count];
  }

  cell = [routeMatrix cellAtRow: count column: 0];   
  [cell setStringValue: rname];
  [cell setLeaf: YES];
  [routeMatrix selectCellAtRow: count column: 0]; 
  [routeMatrix sizeToCells]; 

  [pos addRouteWithName: rname];
}

- (IBAction)delRouteAction:(id)sender
{
  id cell = [routeMatrix selectedCell];
    
  if (cell) {
    [pos removeRouteWithName: [cell stringValue]];
  
    if ([[routeMatrix cells] count] == 1) {
      [routeMatrix removeColumn: 0];
    } else {
      int row, col;
      [routeMatrix getRow: &row column: &col ofCell: cell];
      [routeMatrix removeRow: row];
    }
  
    [routeMatrix sizeToCells]; 
    
    if ([[pointsMatrix cells] count]) {
      [pointsMatrix removeColumn: 0];
      [pointsMatrix sizeToCells]; 
    }
    
    if ([[routeMatrix cells] count] == 0) {
      [follRouteSwtch setState: NSOffState];
      [follRouteSwtch setEnabled: NO];
      [pos stopFollowRoute];
    }
  }    
}

- (IBAction)revRouteAction:(id)sender
{
  id cell = [routeMatrix selectedCell];

  if (cell) {
    NSString *name = [cell stringValue];
    NSArray *points = [pos pointsOfRouteWithName: name];
    
    if (points && [points count]) {
      NSString *revname = [name stringByAppendingString: @"_rev"]; 
      NSMutableArray *newpoints = [NSMutableArray array];
      NSArray *cells = [routeMatrix cells];
      int count = [cells count];
      id cell;
      int i;

      for (i = 0; i < count; i++) {
        if ([[[cells objectAtIndex: i] stringValue] isEqual: revname]) {
          NSRunAlertPanel(NSLocalizedString(@"error", @""), 
                NSLocalizedString(@"A route with this name already exists!", @""), 
                                        NSLocalizedString(@"OK", @""), nil, nil); 
          return;
        }
      }
    
      [routeMatrix insertRow: count];

      cell = [routeMatrix cellAtRow: count column: 0];   
      [cell setStringValue: revname];
      [cell setLeaf: YES];
      [routeMatrix sizeToCells]; 

      [pos addRouteWithName: revname];

      for (i = [points count] - 1; i >= 0; i--) {
        [newpoints insertObject: [points objectAtIndex: i] 
                        atIndex: [newpoints count]];
      }
      
      [pos addWayPoints: newpoints toRouteWithName: revname];
    }
  }
}

- (IBAction)follRouteAction:(id)sender
{
  if ([sender state] == NSOffState) {
    [pos stopFollowRoute];
  } else {
    id cell = [routeMatrix selectedCell];

    if (cell) {
      [pos followRouteWithName: [cell stringValue]];
    }
  }
}

- (void)routeMatrixAction:(id)sender
{
  NSString *rname = [[routeMatrix selectedCell] stringValue];
  NSArray *points = [pos pointsOfRouteWithName: rname];
  NSSize cs, ms;

  [pointsMatrix removeFromSuperview];
	[pointsScroll setDocumentView: nil];	
  DESTROY (pointsMatrix);

  pointsMatrix = [[NSMatrix alloc] initWithFrame: NSMakeRect(0, 0, 100, 100)
				          mode: NSRadioModeMatrix prototype: [NSBrowserCell new]
			       											      numberOfRows: 0 numberOfColumns: 0];
  [pointsMatrix setIntercellSpacing: NSZeroSize];
  [pointsMatrix setCellSize: NSMakeSize(130, 16)];
  [pointsMatrix setAutoscroll: YES];
	[pointsMatrix setAllowsEmptySelection: YES];
  cs = [pointsScroll contentSize];
  ms = [pointsMatrix cellSize];
  ms.width = cs.width;
  CHECKSIZE (ms);
  [pointsMatrix setCellSize: ms];
	[pointsScroll setDocumentView: pointsMatrix];	

  if (points && [points count]) {
    int i;

    [pointsMatrix addColumn];

    for (i = 0; i < [points count]; i++) {
      NSDictionary *dict = [points objectAtIndex: i];
      NSString *pname = [dict objectForKey: @"name"];
      id cell;

      if (i != 0) {
		    [pointsMatrix insertRow: i];
      }     
      cell = [pointsMatrix cellAtRow: i column: 0];   
      [cell setStringValue: pname];
      [cell setLeaf: YES];
    }

    [pointsMatrix sizeToCells]; 
  }

  [routeNameField setStringValue: rname];   
  
  if ([follRouteSwtch isEnabled] == NO) {
    [follRouteSwtch setEnabled: YES];
  }
}

- (IBAction)addPointAction:(id)sender
{
  WayPoint *wpoint = [pos selectedWayPoint];
  
  if (wpoint) {
    NSString *wpname = [wpoint name];
    NSArray *cells = [pointsMatrix cells];
    int count = [cells count];
    id cell;
    int i;
  
    for (i = 0; i < count; i++) {
      if ([[[cells objectAtIndex: i] stringValue] isEqual: wpname]) {
        NSRunAlertPanel(NSLocalizedString(@"error", @""), 
              NSLocalizedString(@"This point already exists!", @""), 
                                      NSLocalizedString(@"OK", @""), nil, nil); 
        return;
      }
    }
    
    if (count == 0) {
      [pointsMatrix addColumn];
    } else {
      [pointsMatrix insertRow: count];
    }

    cell = [pointsMatrix cellAtRow: count column: 0];   
    [cell setStringValue: wpname];
    [cell setLeaf: YES];
    [pointsMatrix selectCellAtRow: count column: 0]; 
    [pointsMatrix sizeToCells]; 

    [pos addWayPoint: wpoint toRouteWithName: [[routeMatrix selectedCell] stringValue]];
  
  } else {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
          NSLocalizedString(@"You must select a waypoint in the map window!", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
  }
}

- (IBAction)delPointAction:(id)sender
{
  id cell = [pointsMatrix selectedCell];
    
  if (cell) {
    [pos removeWayPointWithName: [cell stringValue] 
                inRouteWithName: [[routeMatrix selectedCell] stringValue]];
  
    if ([[pointsMatrix cells] count] == 1) {
      [pointsMatrix removeColumn: 0];
    } else {
      int row, col;
      [pointsMatrix getRow: &row column: &col ofCell: cell];
      [pointsMatrix removeRow: row];
    }
  
    [pointsMatrix sizeToCells]; 
  }    
}

- (void)endFollowRoute:(NSNotification *)notif
{
  NSNumber *n = (NSNumber *)[notif object];
  
  if ([n floatValue] < 0) {
    [follRouteSwtch setState: NSOffState];
  }
}

- (id)inspView
{
  return mainBox;
}

- (void)deactivate
{
  [mainBox removeFromSuperview];
}

- (NSString *)inspname
{
	return NSLocalizedString(@"Routes", @"");
}

- (NSString *)winname
{
	return NSLocalizedString(@"Routes Inspector", @"");
}
	
@end
