/* 
 * SatsInspector.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "SatsInspector.h"
#include "Position.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"

#ifndef max
#define max(a,b) ((a) >= (b) ? (a):(b))
#endif

#ifndef min
#define min(a,b) ((a) <= (b) ? (a):(b))
#endif

#define SATS 12
#define IND_HEIGHT 46.0
#define IND_SPACE 15
#define LABELW 22
#define LABELH 16
#define MAX_STRENGTH 50
#define IM_RADIUS 74.0

static NSString *inspectorNibName = @"SatsInspector";
static NSString *indicatorNibName = @"SatIndicator";

@implementation SatsInspector

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  TEST_RELEASE (indicators);
  TEST_RELEASE (satsView);
  TEST_RELEASE (mainBox);
  
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: inspectorNibName owner: self] == NO) {
      NSLog(@"failed to load %@!", inspectorNibName);
    } else {     
      int i;
      int posx;
       
      RETAIN (mainBox);
      RELEASE (win);
      
      satsView = [[SatsView alloc] init];
      [(NSBox *)mapBox setContentView: satsView];
      
      indicators = [NSMutableArray new];
      
      posx = 0;
      for (i = 0; i < SATS; i++) {
        SatIndicator *indicator = [[SatIndicator alloc] init];
        NSBox *indBox = [indicator mainBox];
        NSRect r = [indBox frame];
        
        [indicatorsBox addSubview: indBox];
        r.origin.x = posx;
        [indBox setFrame: r];
        posx += IND_SPACE;
        
        [indicators addObject: indicator];
        RELEASE (indicator);
      }
      
      usedSats = 0;
      
      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(setUsedSats:) 
                					  name: GPSSetUsedSatsNotification
                					object: nil];

      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(updateSatellites:) 
                					  name: GPSUpdateSatellitesNotification
                					object: nil];
    }
  }
  
  return self;
}

- (void)setUsedSats:(NSNotification *)notif
{
  usedSats = [[notif object] intValue];
}

- (void)updateSatellites:(NSNotification *)notif
{
  NSArray *satsArr = (NSArray *)[notif object];
  int i;
  
  if (usedSats == 0) {
    return;
  }
  
  for (i = 0; i < [satsArr count]; i++) {
    if (i < [indicators count]) {
      NSDictionary *sdict = [satsArr objectAtIndex: i];
      NSString *sname = [sdict objectForKey: @"name"];
      int elevation = [[sdict objectForKey: @"elevation"] intValue];
      int azimuth = [[sdict objectForKey: @"azimuth"] intValue];
      int strength = [[sdict objectForKey: @"strength"] intValue];
      SatIndicator *indicator = [indicators objectAtIndex: i];
       
      [indicator setSatName: sname];
      [indicator setStrength: strength];
      
      [satsView setElevation: elevation 
                     azimuth: azimuth 
              forSatWithName: sname
                     atIndex: i];
                     
      [satsView setStrength: strength forSatAtIndex: i];
    }
  }
}

- (SatIndicator *)indicatorWithName:(NSString *)iname
{
  int i;
  
  for (i = 0; i < [indicators count]; i++) {
    SatIndicator *indicator = [indicators objectAtIndex: i];
    NSString *satName = [indicator satName];
    
    if (satName && [satName isEqual: iname]) {
      return indicator;
    }
  }
  
  return nil;
}

- (id)inspView
{
  return mainBox;
}

- (void)deactivate
{
  [mainBox removeFromSuperview];
}

- (NSString *)inspname
{
	return NSLocalizedString(@"Satellites", @"");
}

- (NSString *)winname
{
	return NSLocalizedString(@"Satellites Inspector", @"");
}
	
@end

@implementation SatsView

- (void)dealloc
{
  RELEASE (image);
  RELEASE (labels);
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
    NSRect r = NSMakeRect(0, 0, LABELW, LABELH);
    int i;
    
    labels = [NSMutableArray new];
  
    for (i = 0; i < SATS; i++) {
      NSTextField *field = [[NSTextField alloc] initWithFrame: r];
    
      [field setEditable: NO];
      [field setSelectable: NO];
      [field setBezeled: NO];
      [field setBordered: NO];
      [field setBackgroundColor: [NSColor blackColor]];
      [field setDrawsBackground: NO];
      [field setAlignment: NSCenterTextAlignment];
      [field setStringValue: @"--"];
      [field setFrame: NSMakeRect(-LABELW, -LABELH, LABELW, LABELH)];
      
      [labels addObject: field];
      [self addSubview: field];
      RELEASE (field);
    }
    
    ASSIGN (image, [NSImage imageNamed: @"satCircle"]);
  }
  
  return self;
}

- (void)setElevation:(int)elevation 
             azimuth:(int)azimuth 
      forSatWithName:(NSString *)name
             atIndex:(int)index
{
  NSRect r = [self frame];
  float centerx = r.size.width / 2;
  float centery = r.size.height / 2;
  float x, y, z;
  
  z = cosdegr(elevation) * IM_RADIUS;
  x = sindegr(azimuth) * z;
  y = cosdegr(azimuth) * z;

  [[labels objectAtIndex: index] setStringValue: name];
  
  [[labels objectAtIndex: index] setFrame: 
        NSMakeRect(centerx + x - LABELH / 2, centery + y - LABELW / 2, LABELW, LABELH)];

  [self setNeedsDisplay: YES];
}

- (void)setStrength:(int)str forSatAtIndex:(int)index
{
  NSTextField *label = [labels objectAtIndex: index];

  if (str > 10) {
    [label setTextColor: [NSColor whiteColor]];
    [label setDrawsBackground: YES];
  } else {
    [label setTextColor: [NSColor blackColor]];
    [label setDrawsBackground: NO];
  }
}

- (void)drawRect:(NSRect)rect
{
  NSSize ssize = [self frame].size;
  NSSize isize = [image size];
  float x = ssize.width / 2 - isize.width / 2;
  float y = ssize.height / 2 - isize.height / 2;

  [image compositeToPoint: NSMakePoint(x, y) 
                operation: NSCompositeSourceOver];
}

@end

@implementation SatIndicator

- (void)dealloc
{
  TEST_RELEASE (mainBox);
  TEST_RELEASE (signIndicator);
  TEST_RELEASE (satName);
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: indicatorNibName owner: self] == NO) {
      NSLog(@"failed to load %@!", indicatorNibName);
    } else {
      NSRect r;
      
      RETAIN (mainBox);
      RELEASE (win);

      signIndicator = [[SignalIndicator alloc] init];
      [(NSBox *)indicatorBox setContentView: signIndicator];
      r = [[(NSBox *)indicatorBox contentView] frame];
      indorx = r.origin.x;
      indory = r.origin.y;
      indw = r.size.width;
      r.size.height = 1;
      [signIndicator setFrame: r];
      
      satName = nil;
      [label setStringValue: @"00"];
      
      strength = 0.0;
      lastStr = 0.0;
    }
  }
  
  return self;
}

- (void)setSatName:(NSString *)sname
{
  if (satName == nil) {
    ASSIGN (satName, sname);
    [label setStringValue: satName];
  }
}

- (NSString *)satName
{
  return satName;
}

- (void)setStrength:(int)str
{
  strength = str;
  
  if (((max(str, lastStr) - min(str, lastStr)) >= 1) || (str == 0)) {
    if (strength != lastStr) {
      NSRect r;
      
      if (strength != 0) {
        r = NSMakeRect(indorx, indory, indw, strength * IND_HEIGHT / MAX_STRENGTH);
      } else {
        r = NSMakeRect(indorx, indory, indw, 1);
      }
      
      [signIndicator setFrame: r];
      [signIndicator setNeedsDisplay: YES];
      [indicatorBox setNeedsDisplay: YES];

      lastStr = strength;
    }
  }  
}
      
- (int)strength
{
  return strength;
}

- (id)label
{
  return label;
}

- (id)mainBox
{
  return mainBox;
}

@end

@implementation SignalIndicator

- (void)dealloc
{
  RELEASE (color);
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
    ASSIGN (color, [NSColor blueColor]);
  }
  
  return self;
}

- (void)drawRect:(NSRect)rect
{
  [color set];
  NSRectFill(rect);
}

@end
