/* 
 * WayPointInspector.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "WayPointInspector.h"
#include "WayPoint.h"
#include "Position.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"

#define CHECKSIZE(sz) \
if (sz.width < 0) sz.width = 0; \
if (sz.height < 0) sz.height = 0

static NSString *inspectorNibName = @"WayPointInspector";

@implementation WayPointInspector

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver: self];
  RELEASE (matrix);
  TEST_RELEASE (mainBox);
  
  [super dealloc];
}

- (id)init
{
  self = [super init];
    
  if(self) {
		if ([NSBundle loadNibNamed: inspectorNibName owner: self] == NO) {
      NSLog(@"failed to load %@!", inspectorNibName);
    } else {   
      NSArray *iconsNames;  
      NSSize cs, ms;
      int i;
         
      RETAIN (mainBox);
      RELEASE (win);
      
      pos = [Position position];
      
      [scrollView setBorderType: NSBezelBorder];
		  [scrollView setHasHorizontalScroller: NO];
  	  [scrollView setHasVerticalScroller: YES]; 

      matrix = [[NSMatrix alloc] initWithFrame: NSMakeRect(0, 0, 100, 100)
				            	mode: NSRadioModeMatrix prototype: [NSBrowserCell new]
			       											          numberOfRows: 0 numberOfColumns: 0];
      [matrix setTarget: self];
      [matrix setAction: @selector(matrixAction:)];
      [matrix setIntercellSpacing: NSZeroSize];
      [matrix setCellSize: NSMakeSize(130, 16)];
      [matrix setAutoscroll: YES];
	    [matrix setAllowsEmptySelection: YES];
      cs = [scrollView contentSize];
      ms = [matrix cellSize];
      ms.width = cs.width;
      CHECKSIZE (ms);
      [matrix setCellSize: ms];
	    [scrollView setDocumentView: matrix];	

      iconsNames = [pos wayPointIconsNames];
      
      [matrix addColumn];   
      for (i = 0; i < [iconsNames count]; i++) {
        NSString *iconName = [iconsNames objectAtIndex: i];
        id cell;

        if (i != 0) {
		      [matrix insertRow: i];
        }     
        cell = [matrix cellAtRow: i column: 0];   
        [cell setStringValue: iconName];
        [cell setLeaf: YES];
      } 
      [matrix sizeToCells]; 

      [matrix setEnabled: NO];
      [applyButt setEnabled: NO];
      
      currentPoint = nil;
      
      [[NSNotificationCenter defaultCenter] addObserver: self 
                				selector: @selector(setWayPoint:) 
                					  name: GPSSelectedWayPointDidChange
                					object: nil];
      
      /* Internationalization */
      [applyButt setTitle: NSLocalizedString(@"Apply", @"")];
    }
  }
  
  return self;
}

- (void)setWayPoint:(NSNotification *)notif
{
  if ([notif object]) {
    WayPoint *wpoint = (WayPoint *)[notif object];
    NSArray *cells = [matrix cells];
    NSString *iconName = [wpoint iconName];
    int i;
    
    [matrix setEnabled: YES];
    [applyButt setEnabled: YES];
    
    for (i = 0; i < [cells count]; i++) {
      id cell = [cells objectAtIndex: i];
    
      if ([[cell stringValue] isEqual: iconName]) {
        int row, col;
        
        [matrix getRow: &row column: &col ofCell: cell];
        [matrix selectCellAtRow: row column: col];
        [matrix sendAction];
        break;
      }
    }
    
    [nameField setTextColor: [NSColor blackColor]];
    [nameField setStringValue: [wpoint name]];
    [infoView setEditable: YES];
    [infoView setString: [wpoint description]];
    currentPoint = wpoint;
    
  } else {
    [nameField setTextColor: [NSColor grayColor]];
    [nameField setStringValue: NSLocalizedString(@"no way point", @"")];
    [infoView setString: @""];
    [infoView setEditable: NO];
    [imageView setImage: nil];
    [matrix setEnabled: NO];
    [applyButt setEnabled: NO];
    currentPoint = nil;
  }
}

- (void)matrixAction:(id)sender
{
  NSString *imgname = [[sender selectedCell] stringValue];

  [imageView setImage: [NSImage imageNamed: imgname]];
}

- (IBAction)applyAction:(id)sender
{
  if (currentPoint) {
    NSMutableDictionary *info = [NSMutableDictionary dictionary];
  
    [info setObject: [currentPoint name] forKey: @"name"];
    [info setObject: [infoView string] forKey: @"description"];
    [info setObject: [[matrix selectedCell] stringValue] forKey: @"iconname"];
    [info setObject: [NSNumber numberWithFloat: [currentPoint longitude]] 
             forKey: @"longitude"];
    [info setObject: [NSNumber numberWithFloat: [currentPoint latitude]] 
             forKey: @"latitude"];
    [info setObject: [NSNumber numberWithFloat: [currentPoint scale]] 
             forKey: @"scale"];

    [currentPoint setWaiPointInfo: info];
  }
}

- (id)inspView
{
  return mainBox;
}

- (void)deactivate
{
  [mainBox removeFromSuperview];
}

- (NSString *)inspname
{
	return NSLocalizedString(@"Waypoint", @"");
}

- (NSString *)winname
{
	return NSLocalizedString(@"Waypoint Inspector", @"");
}
	
@end
