/* 
 * MapView.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include <math.h>
#include "MapView.h"
#include "WayPoint.h"
#include "Position.h"
#include "Functions.h"
#include "GNUstep.h"

#define WPOINTWH 20
#define WPOINTOR (WPOINTWH / 2)
#define LABEL_HEIGHT 14

@implementation MapView

- (void)dealloc
{
  TEST_RELEASE (map);
  RELEASE (pointImg);
  RELEASE (xpoints);
  RELEASE (ypoints);
  RELEASE (wayPoints);
  
  [super dealloc];
}

- (id)init
{
	self = [super initWithFrame: NSZeroRect];
	
	if (self) {
    map = nil;
    mapCenter = NSZeroPoint;
    scale = 0;
    scrPoint = NSZeroPoint;
    ASSIGN (pointImg, [NSImage imageNamed: @"point.tiff"]);
    xpoints = [NSMutableArray new];
    ypoints = [NSMutableArray new];
    tracePath = NO;
    wayPoints = [NSMutableArray new];
    
    position = [Position position];
  }
		
	return self;
}

- (BOOL)setMapPath:(NSString *)path 
         mapCenter:(NSPoint)mcenter 
          mapScale:(long)mscale
         wayPoints:(NSArray *)wpoints
{
  map = [[NSImage alloc] initWithContentsOfFile: path];

  if (map == nil) {
    return NO;
  } else {    
    int i, count;

    mapCenter = mcenter;
    scale = mscale;
        
    border[N] = NSMakeRect(0, MAP_H - BORDER, MAP_W, BORDER);
    border[S] = NSMakeRect(0, 0, MAP_W, BORDER);
    border[E] = NSMakeRect(MAP_W - BORDER, 0, BORDER, MAP_H);
    border[W] = NSMakeRect(0, 0, BORDER, MAP_H);
        
    count = [wayPoints count];
    for(i = 0; i < count; i++) {
      [self removeWayPoint: [wayPoints objectAtIndex: 0]];
    }
   
    if (wpoints) {
      for(i = 0; i < [wpoints count]; i++) {
        [self addWayPoint: [wpoints objectAtIndex: i]];
      }
    }
    
    [self resetTracePath];
  }

  return YES;
}

- (void)setCurrentPosition:(NSPoint)pos checkBorders:(BOOL)chkbords
{
  if (((int)pos.x != (int)scrPoint.x) || ((int)pos.y != (int)scrPoint.y)) {
    scrPoint = NSMakePoint((int)pos.x, (int)pos.y);
    int i;
    
    if (tracePath) {
      int index = [xpoints count];
      
      [xpoints insertObject: [NSNumber numberWithInt: (int)scrPoint.x] 
                    atIndex: index];
      [ypoints insertObject: [NSNumber numberWithInt: (int)scrPoint.y] 
                    atIndex: index];
    }
    
	  [self setNeedsDisplay: YES];
    
    if (chkbords == NO) {
      return;
    }
    
    for (i = 0; i < 4; i++) {
      if (NSPointInRect(scrPoint, border[i])) {
        int direction;
        
        switch(i) {
          case N:
            if (NSPointInRect(scrPoint, border[E])) {
              direction = NE;  
            } else if (NSPointInRect(scrPoint, border[W])) {
              direction = NW;  
            } else {
              direction = N;  
            }
            break;

          case S:
            if (NSPointInRect(scrPoint, border[E])) {
              direction = SE;  
            } else if (NSPointInRect(scrPoint, border[W])) {
              direction = SW;  
            } else {
              direction = S;  
            }            
            break;

          case E:
            if (NSPointInRect(scrPoint, border[N])) {
              direction = NE;  
            } else if (NSPointInRect(scrPoint, border[S])) {
              direction = SE;  
            } else {
              direction = E;  
            }                        
            break;

          case W:
            if (NSPointInRect(scrPoint, border[N])) {
              direction = NW;  
            } else if (NSPointInRect(scrPoint, border[S])) {
              direction = SW;  
            } else {
              direction = W;  
            }                                    
            break;
    
          default:
            break;
        }
        
        [position shouldChangeMap: direction];
        
        break;
      }
    }
  }
}

- (void)setTracePath:(BOOL)value
{
  if (tracePath != value) {
    tracePath = value;
    
    if (tracePath == NO) {
      [xpoints removeAllObjects];
      [ypoints removeAllObjects];
    }
  }
}

- (void)resetTracePath
{
  [xpoints removeAllObjects];
  [ypoints removeAllObjects];
}

- (BOOL)tracePath
{
  return tracePath;
}

- (void)addWayPoint:(NSDictionary *)wdict
{
  WayPoint *wpoint;
  NSPoint ep, sp;

  wpoint = [[WayPoint alloc] initWithDictionary: wdict];

  ep.x = [[wdict objectForKey: @"longitude"] floatValue];
  ep.y = [[wdict objectForKey: @"latitude"] floatValue];

  sp = screenPointForEarthPoint(ep, mapCenter, scale);

  [wpoint setFrame: NSMakeRect(sp.x - WPOINTOR, sp.y - WPOINTOR, WPOINTWH, WPOINTWH)];
  [self setLabelFrameOfWayPoint: wpoint];

  [self addSubview: wpoint];
  [self addSubview: [wpoint label]];

  [wayPoints addObject: wpoint];
  
  RELEASE (wpoint);
}

- (void)removeWayPoint:(WayPoint *)wpoint
{
  [wpoint removeFromSuperview];
  [[wpoint label] removeFromSuperview];
  [wayPoints removeObject: wpoint];
  [self setNeedsDisplay: YES];
}

- (WayPoint *)selectedWayPoint
{
  int i;
  
  for (i = 0; i < [wayPoints count]; i++) {
    WayPoint *wp = [wayPoints objectAtIndex: i];
    
    if ([wp isSelect]) {
      return wp;
    }
  }
  
  return nil;
}

- (void)unselectOtherPoints:(WayPoint *)wpoint
{
  int i;
  
  for (i = 0; i < [wayPoints count]; i++) {
    WayPoint *wp = [wayPoints objectAtIndex: i];
    
    if (wp != wpoint) {
      [wp unselect];
    }
  }
}

- (void)setLabelFrameOfWayPoint:(WayPoint *)wpoint
{
	NSTextField *label = [wpoint label];
	float wpwidth = [wpoint frame].size.width;
  float labwidth = [label frame].size.width;
  float labxpos = [wpoint frame].origin.x;

	if(wpwidth > labwidth) {
		labxpos += ((wpwidth - labwidth) / 2);
	} else {
		labxpos -= ((labwidth - wpwidth) / 2);
	}
	
	[label setFrame: NSMakeRect(labxpos, [wpoint frame].origin.y - LABEL_HEIGHT, 
                                                      labwidth, LABEL_HEIGHT)];
	[label setNeedsDisplay: YES];
  [self setNeedsDisplayInRect: [label frame]];
}

- (void)mouseDown:(NSEvent *)theEvent
{
  NSPoint p, earthp;
  
  p = [theEvent locationInWindow];
  p = [self convertPoint: p fromView: nil];

  earthp = earthPointForScreenPoint(p, mapCenter, scale);
  [position clickOnMap: earthp];

  [position setLongitude: earthp.x latitude: earthp.y];
}

- (void)drawRect:(NSRect)rect
{
  if (map) {
    [map compositeToPoint: NSZeroPoint operation: NSCompositeSourceOver];
  }
  
  if (scrPoint.x != 0 && scrPoint.y != 0) {
    if (tracePath && [xpoints count]) {  
      NSBezierPath *bpath;
      int x, y, i;
      
      bpath = [NSBezierPath bezierPath];
      x = [[xpoints objectAtIndex: 0] intValue];
      y = [[ypoints objectAtIndex: 0] intValue];      
      [bpath setLineWidth: 2];
      [bpath moveToPoint: NSMakePoint(x, y)];
    
      for (i = 1; i < [xpoints count]; i++) {
        x = [[xpoints objectAtIndex: i] intValue];
        y = [[ypoints objectAtIndex: i] intValue];
        [bpath lineToPoint: NSMakePoint(x, y)];
      }
    
      [[NSColor greenColor] set];
      [bpath stroke];
    }
    
    [pointImg compositeToPoint: NSMakePoint(scrPoint.x - 5, scrPoint.y - 5)
                     operation: NSCompositeSourceOver];
  }   
}

- (BOOL)acceptsFirstResponder
{
  return YES;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
	return YES;
}

@end
