/* 
 * Position.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include "Position.h"
#include "GPSInterface.h"
#include "MapWindow.h"
#include "MapView.h"
#include "Map.h"
#include "WayPoint.h"
#include "MapDownload.h"
#include "MapImport.h"
#include "Inspectors/InspectorsController.h"
#include "Preferences/PrefController.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"
#include <math.h>

#ifndef HOST_NAME
  #define HOST_NAME @"localhost"
#endif
#ifndef HOST_PORT
  #define HOST_PORT @"2947"
#endif
#ifndef GPSD_PATH
  #define GPSD_PATH @"/usr/local/bin/gpsd"
#endif
#ifndef GPSD_ARGS
  #define GPSD_ARGS @"-p /dev/ttyS0"
#endif

static Position *position = nil;

@implementation Position

+ (Position *)position
{
	if (position == nil) {
		position = [[Position alloc] init];
	}	
  return position;
}

+ (void)initialize
{
	static BOOL initialized = NO;
	
	if (initialized == YES) {
		return;
  }
	
	initialized = YES;
}

+ (void)registerForServices
{
	NSArray *sendTypes = [NSArray arrayWithObjects: NSStringPboardType, nil];	
	NSArray *returnTypes = [NSArray arrayWithObjects: NSStringPboardType, nil];	
	[NSApp registerServicesMenuSendTypes: sendTypes returnTypes: returnTypes];
}

- (void)dealloc
{
//  [[NSDistributedNotificationCenter defaultCenter] removeObserver: self];
//  [[NSNotificationCenter defaultCenter] removeObserver: self];
  RELEASE (gpsint);
  RELEASE (hostName);
  RELEASE (hostPort);
  RELEASE (gpsdPath);
  RELEASE (gpsdArgs);
  TEST_RELEASE (mapWindow);
  RELEASE (mapsDir);
  RELEASE (mapsDictPath);
  TEST_RELEASE (maps);
  TEST_RELEASE (currMapPath);  
  RELEASE (wpointsPath);
  TEST_RELEASE (wayPointsDict);  
  RELEASE (routesPath);
  TEST_RELEASE (routesDict);  
  TEST_RELEASE (followedRoute);  
  TEST_RELEASE (follRouteName);  
  RELEASE (inspController);
  RELEASE (prefController);
  RELEASE (mapDownload);
  RELEASE (mapImport);
  
	[super dealloc];
}

- (id)init
{
  self = [super init];

  if (self) {
    longitude = 0.0;
    latitude = 0.0;
    lonCorr = 0.0;
    latCorr = 0.0;
    scaleIndex = 0;
    
    mapWindow = nil;
    inspController = nil;
    prefController = nil;   
    mapDownload = nil;
    mapImport = nil;
    
    currMapPath = nil;
    currentMap = nil; 
    
    followedRoute = nil;
    follRouteName = nil;
    followRoute = NO;
    followPntInd = 0;
    nextPntDist = 0;
    nextPointDir = 0;
    nearTarget = NO;
    alertDone = YES;
  }
  
  return self;
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	NSUserDefaults *defaults;
  NSFileManager *fm;
  NSString *home;
  BOOL isdir;
  NSDictionary *mapsDict;
  NSDictionary *dict;
  NSArray *keys;
  BOOL foundDefault;
  id entry;
  int i;

  [isa registerForServices];
  
  home = [NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES) lastObject];
  ASSIGN (mapsDir, [home stringByAppendingPathComponent: @"Position"]);

  fm = [NSFileManager defaultManager];

  if (([fm fileExistsAtPath: mapsDir isDirectory: &isdir] && isdir) == NO) {
    if ([fm createDirectoryAtPath: mapsDir attributes: nil] == NO) {
      NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't create the maps directory! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
      [NSApp terminate: self];
    }
  }

  ASSIGN (wpointsPath, [mapsDir stringByAppendingPathComponent: @"wpoints.plist"]);
  if ([fm fileExistsAtPath: wpointsPath] == NO) {
    if ([[NSDictionary dictionary] writeToFile: wpointsPath atomically: YES] == NO) {
      NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't create the default way points dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
      [NSApp terminate: self];
    }
  }
  
  dict = [NSDictionary dictionaryWithContentsOfFile: wpointsPath];
  
  if (dict == nil) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't read the way points dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
    [fm movePath: wpointsPath toPath: [wpointsPath stringByAppendingString: @"_old"] handler: nil];
    [NSApp terminate: self];
  } else {
    wayPointsDict = [dict mutableCopy];
  }

  ASSIGN (routesPath, [mapsDir stringByAppendingPathComponent: @"routes.plist"]);
  if ([fm fileExistsAtPath: routesPath] == NO) {
    if ([[NSDictionary dictionary] writeToFile: routesPath atomically: YES] == NO) {
      NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't create the default routes dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
      [NSApp terminate: self];
    }
  }
  
  dict = [NSDictionary dictionaryWithContentsOfFile: routesPath];
  
  if (dict == nil) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't read the routes dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
    [fm movePath: routesPath toPath: [routesPath stringByAppendingString: @"_old"] handler: nil];
    [NSApp terminate: self];
  } else {
    routesDict = [dict mutableCopy];
  }

  ASSIGN (mapsDictPath, [mapsDir stringByAppendingPathComponent: @"maps.plist"]);
  if ([fm fileExistsAtPath: mapsDictPath] == NO) {
    if ([[NSDictionary dictionary] writeToFile: mapsDictPath atomically: YES] == NO) {
      NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't create the default map dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil);                                     
      [NSApp terminate: self];
    }
  }

  mapsDict = [NSDictionary dictionaryWithContentsOfFile: mapsDictPath];
  
  if (mapsDict == nil) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"Can't read the map dictionary! Quitting now.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
    [fm movePath: mapsDictPath toPath: [mapsDictPath stringByAppendingString: @"_old"] handler: nil];
    [NSApp terminate: self];
  }

  keys = [mapsDict allKeys];

  maps = [NSMutableArray new];

  foundDefault = NO;
  for (i = 0; i < [keys count]; i++) {
    NSString *impath = [keys objectAtIndex: i];

    if ([fm fileExistsAtPath: impath]) {
      NSDictionary *dict = [mapsDict objectForKey: impath];
      float lon = [[dict objectForKey: @"longitude"] floatValue];
      float lat = [[dict objectForKey: @"latitude"] floatValue];
      long scale = [[dict objectForKey: @"scale"] intValue];
      BOOL isdefmap = [[dict objectForKey: @"defaultmap"] boolValue];
      
      if (isdefmap) {
        foundDefault = YES;
      }
      
      Map *map = [[Map alloc] initWithPath: impath longitude: lon 
                               latitude: lat scale: scale isDefMap: isdefmap];
      [maps addObject: map];
      RELEASE (map);
    }
  }

  if (foundDefault == NO) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
             NSLocalizedString(@"No default map!\n\
You must choose a default map from the preferences.", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
  }

  defaults = [NSUserDefaults standardUserDefaults];

  entry = [defaults objectForKey: @"mappath"];
  if (entry && [fm fileExistsAtPath: entry]) {
    ASSIGN (currMapPath, entry);
    
    currentMap = [self mapWithPath: currMapPath];
    
    for (i = 0; i < SCALES; i++) {
      if (scales[i] == [currentMap scale]) {
        scaleIndex = i;
        break;
      }
    }
  } 

  if (currentMap) {
    [self createMapWindow];
  }

  entry = [defaults objectForKey: @"loncorr"];
  if (entry) {
    lonCorr = [entry floatValue];
  }
  entry = [defaults objectForKey: @"latcorr"];
  if (entry) {
    latCorr = [entry floatValue];
  }

  useTuning = [defaults boolForKey: @"usetuning"];
  
  entry = [defaults objectForKey: @"units"];
  if (entry) {
    units = [entry intValue];
  } else {
    units = METRIC;
  }
  
  entry = [defaults objectForKey: @"hostname"];
  if (entry) {
    hostName = RETAIN (entry);
  } else {
    hostName = [[NSString alloc] initWithString: HOST_NAME];
  }
  
  entry = [defaults objectForKey: @"hostport"];
  if (entry) {
    hostPort = RETAIN (entry);
  } else {
    hostPort = [[NSString alloc] initWithString: HOST_PORT];
  }

  entry = [defaults objectForKey: @"gpsdpath"];
  if (entry) {
    gpsdPath = RETAIN (entry);
  } else {
    gpsdPath = [[NSString alloc] initWithString: GPSD_PATH];
  }

  entry = [defaults objectForKey: @"gpsdargs"];
  if (entry) {
    gpsdArgs = RETAIN (entry);
  } else {
    gpsdArgs = [[NSString alloc] initWithString: GPSD_ARGS];
  }
  
  voiceAlerts = [defaults boolForKey: @"sayalerts"];

  inspController = [[InspectorsController alloc] init];
  prefController = [[PrefController alloc] init];
  mapDownload = [[MapDownload alloc] init];
  mapImport = [[MapImport alloc] init];  
  
  [self createGpsInterface];
}

- (void)createGpsInterface
{
  if (gpsint) {
    [gpsint closeConnection];  
    DESTROY (gpsint);  
  }

  gpsint = [[GPSInterface alloc] init];
  if ([gpsint openGpsdConnectionToHost: hostName 
                                atPort: hostPort
                              gpsdPath: gpsdPath
                              gpsdArgs: gpsdArgs] == NO) {
    NSRunAlertPanel(NSLocalizedString(@"error", @""), 
         NSLocalizedString(@"Can't open the gpsd connection! Change the host settings.", @""), 
                                   NSLocalizedString(@"OK", @""), nil, nil);                                     
    DESTROY (gpsint);
  }
}

- (void)gpsdConnectionLost
{
  NSRunAlertPanel(NSLocalizedString(@"error", @""), 
        NSLocalizedString(@"gpsd connection lost or GPS receiver disconnected!", @""), 
                      NSLocalizedString(@"OK", @""), nil, nil);                                     
  DESTROY (gpsint); 
}

- (void)createMapWindow
{
  float lon = [currentMap longitude];
  float lat = [currentMap latitude];
  NSArray *wpoints = [self wayPointsForCenterPoint: NSMakePoint(lon, lat)
                                           atScale: scales[scaleIndex]];

  if (mapWindow) {
    if ([[mapWindow myWin] isVisible]) {
      [[mapWindow myWin] close];
    }
    DESTROY (mapWindow);
  }
  
  mapWindow = [[MapWindow alloc] init];
  [mapWindow activate];

  [[mapWindow mapView] setMapPath: currMapPath
                        mapCenter: NSMakePoint(lon, lat)
                         mapScale: [currentMap scale]
                        wayPoints: wpoints];
                         
  [[mapWindow mapView] setTracePath: YES];
  
  [self setLongitude: lon latitude: lat];
}

- (BOOL)addMapFromFile:(NSString *)path 
             longitude:(float)lon
              latitude:(float)lat
                 scale:(long)scale
{
  NSString *fileName = [path lastPathComponent];
  NSString *filePath = [mapsDir stringByAppendingPathComponent: fileName];

  if ([[NSFileManager defaultManager] copyPath: path toPath: filePath handler: nil]) {
    NSMutableDictionary *mapsDict;
    NSMutableDictionary *newMapDict;
    Map *map;

    map = [[Map alloc] initWithPath: filePath longitude: lon 
                              latitude: lat scale: scale isDefMap: NO];
    [maps addObject: map];
    RELEASE (map);
  
    newMapDict = [NSMutableDictionary dictionary];
    [newMapDict setObject: [NSNumber numberWithFloat: lon] forKey: @"longitude"];
    [newMapDict setObject: [NSNumber numberWithFloat: lat] forKey: @"latitude"];
    [newMapDict setObject: [NSNumber numberWithLong: scale] forKey: @"scale"];
    [newMapDict setObject: [NSNumber numberWithBool: NO] forKey: @"defaultmap"];

    mapsDict = [[NSDictionary dictionaryWithContentsOfFile: mapsDictPath] mutableCopy];
    [mapsDict setObject: newMapDict forKey: fileName];
    [mapsDict writeToFile: mapsDictPath atomically: YES];
    RELEASE (mapsDict);
    
    return YES;
  }
  
  return NO;
}

- (void)setDefaultMapPath:(NSString *)mpath 
                 withInfo:(NSDictionary *)mapinfo
{
  NSMutableDictionary *mapsDict;
  NSArray *keys;
  float lon, lat;
  long scale;
  Map *map; 
  int i;
  
  mapsDict = [[NSDictionary dictionaryWithContentsOfFile: mapsDictPath] mutableCopy];
  keys = [mapsDict allKeys];
  
  for (i = 0; i < [keys count]; i++) {
    NSString *p = [keys objectAtIndex: i];
    NSDictionary *mdict = [mapsDict objectForKey: p];
    BOOL isDefMap = [[mdict objectForKey: @"defaultmap"] boolValue];
    
    if (isDefMap) {
      [mapsDict removeObjectForKey: p];
    }
  }

  [mapsDict setObject: mapinfo forKey: mpath];
  [mapsDict writeToFile: mapsDictPath atomically: YES];
  RELEASE (mapsDict);
  
  lon = [[mapinfo objectForKey: @"longitude"] floatValue];
  lat = [[mapinfo objectForKey: @"latitude"] floatValue];
  scale = [[mapinfo objectForKey: @"scale"] intValue];
  map = [[Map alloc] initWithPath: mpath longitude: lon 
                              latitude: lat scale: scale isDefMap: YES];
  [maps addObject: map];
  RELEASE (map);

  if (currentMap == nil) {
    ASSIGN (currMapPath, mpath);
    currentMap = map;
    
    for (i = 0; i < SCALES; i++) {
      if (scales[i] == [map scale]) {
        scaleIndex = i;
        break;
      }
    }
  
    [self createMapWindow];
  }
}

- (NSString *)defaultMapPath
{
  int i;
  
  for (i = 0; i < [maps count]; i++) {
    Map *map = [maps objectAtIndex: i];
  
    if ([map isDefaultMap]) {
      return [map path];
    }
  }

  return nil;
}

- (void)setLongitude:(float)lon latitude:(float)lat
{
  if (currentMap == nil) {
    return;
  }

  if (lat != 0 && lon != 0) {
    NSPoint earthp, scrp, mapcenter;
  
    printf("longitude = %f latitude = %f\n", lon, lat);
    
    longitude = lon;
    latitude = lat;

    if (useTuning) {
      longitude += lonCorr;
      latitude += latCorr;
    }
  
    earthp = NSMakePoint(longitude, latitude);
    mapcenter = [currentMap center];
    scrp = screenPointForEarthPoint(earthp, mapcenter, scales[scaleIndex]);
  
    [[mapWindow mapView] setCurrentPosition: scrp checkBorders: YES]; 
    
    [inspController setLongitude: longitude latitude: latitude];   
    
    if (followRoute) {
      [self distanceFromNextPoint];
    }
  }
}

- (void)setAltitude:(float)alt
{
  [inspController setAltitude: alt];
}

- (void)setSpeed:(float)speed
{
  [inspController setSpeed: speed];
}

- (void)setTrack:(float)track
{
  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSSetTrackNotification 
                        object: [NSNumber numberWithFloat: track]];
}

- (void)setGpsDate:(NSCalendarDate *)date
{
  int h = [date hourOfDay];
  int m = [date minuteOfHour];
  int s = [date secondOfMinute];
  int zoneOffset = 3;
  NSString *gpsTime = [NSString stringWithFormat: @"%i:%i:%i", h + zoneOffset, m, s];

  [inspController setGpsTime: gpsTime];
}

- (void)setUsedSats:(int)satsnum
{
  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSSetUsedSatsNotification 
                        object: [NSNumber numberWithInt: satsnum]];
}

- (void)updateSatellites:(NSArray *)satsArr
{
  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSUpdateSatellitesNotification 
                        object: satsArr];
}

- (void)clickOnMap:(NSPoint)p
{
  [mapDownload updateLongitude: p.x latitude: p.y];
}

- (void)shouldChangeMap:(int)direction
{
  long scl = scales[scaleIndex];
  NSArray *mps = [self mapsWithScale: scl];
  int i;

  if (mps) {
    for (i = 0; i < [mps count]; i++) { 
      Map *map = [mps objectAtIndex: i]; 
      
      if ([map canDisplayEarthPoint: NSMakePoint(longitude, latitude)] 
                                                    && (map != currentMap)) { 
        NSPoint earthp = NSMakePoint(longitude, latitude); 
        NSPoint mapcenter = [map center]; 
        NSPoint scrp = screenPointForEarthPoint(earthp, mapcenter, scales[scaleIndex]); 
        NSArray *wpoints = [self wayPointsForCenterPoint: mapcenter atScale: scl];
        
        currentMap = map; 
        ASSIGN (currMapPath, [map path]); 
        
        [[mapWindow mapView] setMapPath: currMapPath 
                              mapCenter: mapcenter 
                               mapScale: [map scale] 
                              wayPoints: wpoints]; 

        [[mapWindow mapView] setTracePath: YES]; 
        
        [[mapWindow mapView] setCurrentPosition: scrp checkBorders: NO]; 
        
        return; 
      } 
    }
  }
    
  [self zoom: ZOOMOUT];
}

- (void)zoom:(int)d
{
  NSArray *mps;
  long scl;
  int index;
  NSPoint p;
  int i;

  index = scaleIndex;
  
  if ((d == ZOOMOUT) && (index == (SCALES - 1))) {
    return;
  }
  if ((d == ZOOMIN) && (index == 1)) {
    return;
  }

  if (d == ZOOMOUT) {
    index++;
  } else {
    index--;
  }
  scl = scales[index];

  while (index < SCALES && index > 0) {
    BOOL found = NO;
    
    mps = [self mapsWithScale: scl];
    
    if (mps) {
      for (i = 0; i < [mps count]; i++) {
        Map *map = [mps objectAtIndex: i];

        p = NSMakePoint(longitude, latitude);
        
        if ([map canDisplayEarthPoint: p] && (map != currentMap)) {
          NSArray *wpoints;
          
          scaleIndex = index; 

          currentMap = map;
          ASSIGN (currMapPath, [map path]); 
                    
          wpoints = [self wayPointsForCenterPoint: [map center] 
                                          atScale: [map scale]];
          
          [[mapWindow mapView] setMapPath: currMapPath 
                                mapCenter: [map center]
                                 mapScale: [map scale]
                                wayPoints: wpoints]; 
                                
          [[mapWindow mapView] setTracePath: YES]; 
          
          p = screenPointForEarthPoint(NSMakePoint(longitude, latitude), 
                                     [currentMap center], scales[scaleIndex]);
          [[mapWindow mapView] setCurrentPosition: p checkBorders: NO];

          return;
        }
      }
      
      if (found) {
        break;
      }
    }
    
    if (d == ZOOMOUT) {
      index++;
    } else {
      index--;
    }
    scl = scales[index];
  }
}

- (void)zoomIn:(id)sender
{
  [self zoom: ZOOMIN];
}

- (void)zoomOut:(id)sender
{
  [self zoom: ZOOMOUT];
}

- (Map *)mapWithPath:(NSString *)mpath
{
  int i;
  
  for (i = 0; i < [maps count]; i++) {
    Map *map = [maps objectAtIndex: i];
  
    if ([[map path] isEqual: mpath]) {
      return map;
    }
  }
  
  return nil;
}

- (NSArray *)mapsWithScale:(long)sc
{
  NSMutableArray *mparr = [NSMutableArray array];
  int i;
  
  for (i = 0; i < [maps count]; i++) {
    Map *map = [maps objectAtIndex: i];
  
    if ([map scale] == sc) {
      [mparr addObject: map];
    }
  }
  
  return ([mparr count] ? mparr : nil);
}

- (void)tuneLongitude:(float)value
{
  lonCorr = value;
}

- (float)longitudeTune
{
  return lonCorr;
}

- (void)tuneLatitude:(float)value
{
  latCorr = value;
}

- (float)latitudeTune
{
  return latCorr;
}

- (void)setUseTuning:(BOOL)value
{
  useTuning = value;
  [self setLongitude: longitude latitude: latitude];
}

- (BOOL)useTuning
{
  return useTuning;
}

- (void)setUnits:(int)value
{
  units = value;
  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSDidChangeUnitsNotification object: nil];
}

- (int)units
{
  return units;
}

- (void)connSettingsdidChangeHost:(NSString *)hostname 
                             port:(NSString *)portname
                         gpsdPath:(NSString *)gpsdpath
                         gpsdArgs:(NSString *)gpsdargs
{
  if (([hostname isEqual: hostName] == NO)
                    || ([hostPort isEqual: portname] == NO)
                        || ([gpsdPath isEqual: gpsdpath] == NO)
                            || ([gpsdArgs isEqual: gpsdargs] == NO)) {
    ASSIGN (hostName, hostname);
    ASSIGN (hostPort, portname);
    ASSIGN (gpsdPath, gpsdpath);
    ASSIGN (gpsdArgs, gpsdargs);
    [self createGpsInterface];
  }
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app 
{
#define TEST_CLOSE(o, w) if ((o) && ([w isVisible])) [w close]

  [self updateDefaults];
  
	TEST_CLOSE (inspController, [inspController myWin]);
	TEST_CLOSE (prefController, [prefController myWin]);
	TEST_CLOSE (mapDownload, [mapDownload myWin]);
	TEST_CLOSE (mapImport, [mapImport myWin]);
  
  if (gpsint) {
    [gpsint closeConnection];    
  }
   
	return NSTerminateNow;
}

- (void)updateDefaults
{
  NSMutableDictionary *mapsDict = [NSMutableDictionary dictionary];
  NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
  int i;

  for (i = 0; i < [maps count]; i++) {
    Map *map = [maps objectAtIndex: i];
    NSString *mpath = [map path];
    NSMutableDictionary *dict = [NSMutableDictionary dictionary];

    [dict setObject: [NSNumber numberWithFloat: [map longitude]] 
                                        forKey: @"longitude"];
    [dict setObject: [NSNumber numberWithFloat: [map latitude]] 
                                        forKey: @"latitude"];
    [dict setObject: [NSNumber numberWithLong: [map scale]] 
                                       forKey: @"scale"];
    [dict setObject: [NSNumber numberWithBool: [map isDefaultMap]] 
                                       forKey: @"defaultmap"];
    
    [mapsDict setObject: dict forKey: mpath];
  }

  [mapsDict writeToFile: mapsDictPath atomically: YES];
  
  if (currentMap) {
    [defaults setObject: [currentMap path] forKey: @"mappath"];
  }
  
  [defaults setObject: [NSString stringWithFormat: @"%f", lonCorr] 
               forKey: @"loncorr"];
  [defaults setObject: [NSString stringWithFormat: @"%f", latCorr] 
               forKey: @"latcorr"];
  [defaults setBool: useTuning forKey: @"usetuning"];
  [defaults setObject: [NSNumber numberWithInt: units] forKey: @"units"];

  [defaults setObject: hostName forKey: @"hostname"];
  [defaults setObject: hostPort forKey: @"hostport"];
  [defaults setObject: gpsdPath forKey: @"gpsdpath"];
  [defaults setObject: gpsdArgs forKey: @"gpsdargs"];

  [defaults setBool: voiceAlerts forKey: @"sayalerts"];

  [wayPointsDict writeToFile: wpointsPath atomically: YES];

  [routesDict writeToFile: routesPath atomically: YES];

  [defaults synchronize];
  
  [mapWindow saveDefaults];

	if ((inspController != nil) && ([[inspController myWin] isVisible])) {  
		[inspController updateDefaults]; 
	}

	if ([[prefController myWin] isVisible]) {  
		[prefController updateDefaults]; 
	}

	if ([[mapDownload myWin] isVisible]) {  
		[mapDownload updateDefaults]; 
	}

	if ([[mapImport myWin] isVisible]) {  
		[mapImport updateDefaults]; 
	}
}

- (void)connectToGpsdHost:(id)sender
{
  [self createGpsInterface];
}

- (void)disconnectFromGpsdHost:(id)sender
{
  if (gpsint) {
    [gpsint closeConnection];
    DESTROY (gpsint);
  }
}

- (void)showInspectors:(id)sender
{
  [[inspController myWin] orderFrontRegardless]; 
}

- (void)showSatInspector:(id)sender
{
  [self showInspectors: nil];
  [inspController showSatellites];
}

- (void)showNavigationInspector:(id)sender
{
  [self showInspectors: nil];
  [inspController showNavigation];
}

- (void)showWPointInspector:(id)sender
{
  [self showInspectors: nil];
  [inspController showWayPoints];
}

- (void)showRouteInspector:(id)sender
{
  [self showInspectors: nil];
  [inspController showRoutes];
}

- (void)showPreferences:(id)sender
{
  [prefController activate];
}

- (void)showMapDownload:(id)sender
{
  if (currentMap) {
    NSMutableDictionary *dict = [NSMutableDictionary dictionary];

    [dict setObject: [NSNumber numberWithFloat: longitude] forKey: @"longitude"];
    [dict setObject: [NSNumber numberWithFloat: latitude] forKey: @"latitude"];
    [dict setObject: [NSNumber numberWithLong: scales[scaleIndex]] forKey: @"scale"];

    [mapDownload activateWithInfo: dict];
  }
}

- (void)showMapImport:(id)sender
{
  [mapImport setFromLocalBehaviour];
  [mapImport activate];
}

- (void)showMapImportForDownloadedFile:(NSDictionary *)fileInfo
{
  [mapImport setFromNetBehaviour];
  [mapImport activate];
  [mapImport prepareToImportDownloadedFile: fileInfo];
}

- (BOOL)validateMenuItem:(NSMenuItem *)anItem 
{	
	NSString *title = [anItem title];
	
	if ([title isEqual: NSLocalizedString(@"Remove way point", @"")]) {
    if (mapWindow && ([[mapWindow mapView] selectedWayPoint] != nil)) {
      return YES;
		} else {
      return NO;
    }
	} 

	if ([title isEqual: NSLocalizedString(@"Next point in route", @"")]) {
    if (followRoute && followedRoute && ([followedRoute count] > (followPntInd + 1))) {
      return YES;
    } else {
      return NO;
    }
	} 

	if ([title isEqual: NSLocalizedString(@"Connect to gpsd", @"")]) {
    if (gpsint) { 
      return NO;
    }
  }

  if ([title isEqual: NSLocalizedString(@"Disconnect from gpsd", @"")]) {
    if (gpsint == nil) { 
      return NO;
    }
  }
    
	return YES;
}

- (void)showInfo:(id)sender
{
  NSMutableDictionary *d = AUTORELEASE ([NSMutableDictionary new]);
  [d setObject: @"Position" forKey: @"ApplicationName"];
  [d setObject: NSLocalizedString(@"GNUstep earth navigation system", @"")
      	forKey: @"ApplicationDescription"];
  [d setObject: @"Position 0.3" forKey: @"ApplicationRelease"];
  [d setObject: @"08 2003" forKey: @"FullVersionID"];
  [d setObject: [NSArray arrayWithObjects: 
      @"Enrico Sersale <enrico@imago.ro>", nil]
     forKey: @"Authors"];
  [d setObject: NSLocalizedString(@"See http://www.gnustep.it/enrico/gps", @"") forKey: @"URL"];
  [d setObject: @"Copyright (C) 2003 Enrico Sersale."
     forKey: @"Copyright"];
  [d setObject: NSLocalizedString(@"Released under the GNU General Public License 2.0", @"")
     forKey: @"CopyrightDescription"];
  
#ifdef GNUSTEP	
  [NSApp orderFrontStandardInfoPanelWithOptions: d];
#else
	[NSApp orderFrontStandardAboutPanel: d];
#endif
}

#ifndef GNUSTEP
- (void)terminate:(id)sender
{
  [NSApp terminate: self];
}
#endif

@end



