/* 
 * DefaultMapPref.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include "DefaultMapPref.h"
#include "Functions.h"
#include "Position.h"
#include "GNUstep.h"

static NSString *nibName = @"DefaultMapPref";

#define CHECKSIZE(sz) \
if (sz.width < 0) sz.width = 0; \
if (sz.height < 0) sz.height = 0

@implementation DefaultMapPref

- (void)dealloc
{
  TEST_RELEASE (prefbox);
  TEST_RELEASE (matrix);
  RELEASE (cellPrototipe); 
  TEST_RELEASE (resPath);
  TEST_RELEASE (mapsDict);
  TEST_RELEASE (defMapPath);
  [super dealloc];
}

- (id)init
{
  self = [super init];
  
  if (self) {
		if ([NSBundle loadNibNamed: nibName owner: self] == NO) {
      NSLog(@"failed to load %@!", nibName);
    } else {
      RETAIN (prefbox);
      RELEASE (win); 

      pos = [Position position];   
		  fm = [NSFileManager defaultManager];
      
      cellPrototipe = [NSBrowserCell new];

      [scrollView setBorderType: NSBezelBorder];
		  [scrollView setHasHorizontalScroller: NO];
  	  [scrollView setHasVerticalScroller: YES]; 

      matrix = nil;
      
      /* Internationalization */
      [okButt setTitle: NSLocalizedString(@"Ok", @"")];
      [labelInfo setStringValue: NSLocalizedString(@"Choose your default map", @"")];

      [self setContents];
    }
  }
  
  return self;
}

- (void)setContents
{
  NSString *dictPath;
  NSArray *mapsNames;
  NSString *mapPath;
  NSSize cs, ms;
  int i;

  defMapPath = nil;
  
  ASSIGN (resPath, [[NSBundle mainBundle] resourcePath]); 

#ifdef GNUSTEP 
  dictPath = [resPath stringByAppendingPathComponent: @"defaultmaps.plist"];
#else
  dictPath = [resPath stringByAppendingPathComponent: @"defaultmaps_mac.plist"];
#endif
  
  mapsDict = [NSDictionary dictionaryWithContentsOfFile: dictPath];
  if (mapsDict == nil) {
    return;
  }
  
  RETAIN (mapsDict);
    
  matrix = [[NSMatrix alloc] initWithFrame: NSMakeRect(0, 0, 100, 100)
				            			  mode: NSRadioModeMatrix prototype: cellPrototipe
			       											          numberOfRows: 0 numberOfColumns: 0];
  [matrix setTarget: self];
  [matrix setAction: @selector(matrixAction:)];
  [matrix setIntercellSpacing: NSZeroSize];
  [matrix setCellSize: NSMakeSize(130, 16)];
  [matrix setAutoscroll: YES];
	[matrix setAllowsEmptySelection: YES];
  cs = [scrollView contentSize];
  ms = [matrix cellSize];
  ms.width = cs.width;
  CHECKSIZE (ms);
  [matrix setCellSize: ms];
	[scrollView setDocumentView: matrix];	
       
  mapsNames = [mapsDict allKeys];  
    
  [matrix addColumn];   
  for (i = 0; i < [mapsNames count]; i++) {
    NSString *mapName = [mapsNames objectAtIndex: i];
    NSString *mapPath = [resPath stringByAppendingPathComponent: mapName];
  
    if ([fm fileExistsAtPath: mapPath]) {
      id cell;
		
      if (i != 0) {
		    [matrix insertRow: i];
      }     
      cell = [matrix cellAtRow: i column: 0];   
      [cell setStringValue: mapName];
      [cell setLeaf: YES];
    } 
  } 
  [matrix sizeToCells]; 
  
  mapPath = [pos defaultMapPath];
  
  if (mapPath) {
    NSString *mapName = [mapPath lastPathComponent];
    NSArray *cells = [matrix cells];
    int i;
    
    [matrix deselectAllCells];
  
	  for (i = 0; i < [cells count]; i++) {
	    NSCell *cell = [cells objectAtIndex: i];
    
      if ([[cell stringValue] isEqual: mapName]) {
        [matrix selectCell: cell];
        [matrix sendAction];
      } 
	  }
  }
}

- (void)matrixAction:(id)sender
{
  NSString *mapName = [[matrix selectedCell] stringValue];
  NSDictionary *dict = [mapsDict objectForKey: mapName];
#ifdef GNUSTEP 
  NSNumber *lon = [dict objectForKey: @"longitude"];
  NSNumber *lat = [dict objectForKey: @"latitude"];
#else
  NSString *lon = [dict objectForKey: @"longitude"];
  NSString *lat = [dict objectForKey: @"latitude"];
#endif
  NSString *mapPath = [resPath stringByAppendingPathComponent: mapName];
  NSImage *map = [[NSImage alloc] initWithContentsOfFile: mapPath];
  
  [imageView setImage: map];
  RELEASE (map);
  
#ifdef GNUSTEP 
  [lonField setStringValue: [lon stringValue]];
  [latField setStringValue: [lat stringValue]];
#else
  [lonField setStringValue: lon];
  [latField setStringValue: lat];
#endif
  
  ASSIGN (defMapPath, mapPath);
}

- (IBAction)okAction:(id)sender
{
  if (defMapPath) {
    NSString *mapName = [[matrix selectedCell] stringValue];
    NSDictionary *mdict = [mapsDict objectForKey: mapName];
  
    [pos setDefaultMapPath: defMapPath withInfo: mdict];
  }
}

- (NSView *)prefView
{
  return prefbox;
}

- (NSString *)prefName
{
  return NSLocalizedString(@"Default map", @"");
}

@end
