/* 
 * Routes.m
 * 
 * Copyright (c) 2003 Enrico Sersale <enrico@imago.ro>
 *
 * Author: Enrico Sersale <enrico@imago.ro>
 * Date: August 2003
 *
 * This file is part of the GNUstep Position application
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <Foundation/Foundation.h>
#include <AppKit/AppKit.h>
#include "Position.h"
#include "GPSInterface.h"
#include "MapWindow.h"
#include "MapView.h"
#include "Map.h"
#include "WayPoint.h"
#include "MapDownload.h"
#include "MapImport.h"
#include "Inspectors/InspectorsController.h"
#include "Preferences/PrefController.h"
#include "Functions.h"
#include "Notifications.h"
#include "GNUstep.h"

#define NEAR_POINT (0.2)
#define TARGET (0.03)

@implementation Position (Routes)

- (NSDictionary *)routesDict
{
  return routesDict;
}

- (void)addRouteWithName:(NSString *)rname
{
  [routesDict setObject: [NSArray array] forKey: rname];
}

- (void)removeRouteWithName:(NSString *)rname
{
  [routesDict removeObjectForKey: rname];
}

- (void)addWayPoint:(WayPoint *)wpoint toRouteWithName:(NSString *)rname
{
  NSMutableArray *pnts = [[routesDict objectForKey: rname] mutableCopy];
  NSMutableDictionary *wpdict = [NSMutableDictionary dictionary];
  int index = [pnts count];
  
  [wpdict setObject: [wpoint name] forKey: @"name"];
  [wpdict setObject: [NSNumber numberWithFloat: index] forKey: @"index"];
  [wpdict setObject: [wpoint description] forKey: @"description"];
  [wpdict setObject: [wpoint iconName] forKey: @"iconname"];
  [wpdict setObject: [NSNumber numberWithFloat: [wpoint longitude]] 
             forKey: @"longitude"];
  [wpdict setObject: [NSNumber numberWithFloat: [wpoint latitude]] 
             forKey: @"latitude"];
  [wpdict setObject: [NSNumber numberWithLong: [wpoint scale]] 
             forKey: @"scale"];
  
  [pnts insertObject: wpdict atIndex: [pnts count]];
  [routesDict setObject: pnts forKey: rname];
  RELEASE (pnts);
}

- (void)addWayPoints:(NSArray *)points toRouteWithName:(NSString *)rname
{
  [routesDict setObject: points forKey: rname];
}

- (void)removeWayPointWithName:(NSString *)wpname
               inRouteWithName:(NSString *)rname
{
  NSMutableArray *pnts = [[routesDict objectForKey: rname] mutableCopy];
  int count = [pnts count];
  int i;

  for (i = 0; i < count; i++) {
    NSDictionary *wpdict = [pnts objectAtIndex: i];
    NSString *name = [wpdict objectForKey: @"name"];
  
    if ([name isEqual: wpname]) {
      [pnts removeObjectAtIndex: i];
      count--;
      i--;
    }
  }

  [routesDict setObject: pnts forKey: rname];
  RELEASE (pnts);
}

- (NSArray *)pointsOfRouteWithName:(NSString *)rname
{
  return [routesDict objectForKey: rname];
}

- (NSDictionary *)infoOfWayPointAtIndex:(int)index
{
  int i;
  
  for (i = 0; i < [followedRoute count]; i++) {
    NSDictionary *ptinfo = [followedRoute objectAtIndex: i];
    int ptind = [[ptinfo objectForKey: @"index"] intValue];
  
    if (ptind == index) {
      return ptinfo;
    }
  }

  return nil;
}

- (void)followRouteWithName:(NSString *)rname
{
  ASSIGN (follRouteName, rname);
  ASSIGNCOPY (followedRoute, [routesDict objectForKey: rname]);
  followRoute = YES;
  followPntInd = 0;
  nextPntDist = 0;
  nextPointDir = 0;
  nearTarget = NO;
  alertDone = NO;
}

- (void)stopFollowRoute
{
  followRoute = NO;
  DESTROY (follRouteName);
  DESTROY (followedRoute);
  followPntInd = 0;
  nextPntDist = 0;
  nextPointDir = 0;
  nearTarget = NO;
  alertDone = NO;
  
  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSSetGotoDirNotification 
                        object: [NSNumber numberWithFloat: -1.0]];
}

- (void)nextPointInRoute:(id)sender
{
  if ([followedRoute count] > (followPntInd + 1)) { 
    followPntInd++;
    nextPntDist = 0;
    nextPointDir = 0;
    nearTarget = NO;
    alertDone = NO;
    [self distanceFromNextPoint];
  }
}

- (BOOL)followRoute
{
  return followRoute;
}

- (NSArray *)followedRoute
{
  return followedRoute;
}
  
- (NSString *)follRouteName
{
  return follRouteName;
}

- (void)distanceFromNextPoint
{
  NSDictionary *ptinfo = [self infoOfWayPointAtIndex: followPntInd];
  NSString *ptdesc = [ptinfo objectForKey: @"description"];
  float lon = [[ptinfo objectForKey: @"longitude"] floatValue];
  float lat = [[ptinfo objectForKey: @"latitude"] floatValue];
  NSPoint actp = NSMakePoint(longitude, latitude);
  NSPoint dstp = NSMakePoint(lon, lat);

  nextPntDist = distanceBetweenPoints(actp, dstp);

  [inspController setDistToTarget: nextPntDist];

  nextPointDir = angleBetweenPoints(actp, dstp);  

  [[NSNotificationCenter defaultCenter] 
          postNotificationName: GPSSetGotoDirNotification 
                        object: [NSNumber numberWithFloat: nextPointDir]];

  if (nextPntDist <= NEAR_POINT) {
    nearTarget = YES;  
  
    if (alertDone == NO) {
      [self sayAlert: [NSString stringWithFormat: @"near %@.", ptdesc]];
      alertDone = YES;
    }
    
    if (nextPntDist <= TARGET) {
      if ([followedRoute count] == (followPntInd + 1)) {
        [self stopFollowRoute];
        [self sayAlert: [NSString stringWithFormat: @"arrived to %@.", ptdesc]];
      
        NSRunAlertPanel(NSLocalizedString(@"route", @""), 
                            NSLocalizedString(@"Arrived!", @""), 
                                NSLocalizedString(@"OK", @""), nil, nil); 
      } else {                 
        [self sayAlert: [NSString stringWithFormat: @"arrived to %@. Following next point.", ptdesc]];
        followPntInd++;
        nearTarget = NO;
      }
    }
    
  } else {
    alertDone = NO;

    if (nearTarget) {
      if ([followedRoute count] > (followPntInd + 1)) {        
        [self sayAlert: @"Current waypoint not reached. Following next point"];
        followPntInd++;
        nearTarget = NO;
      } else {
        [self stopFollowRoute];  
        [self sayAlert: [NSString stringWithFormat: @"you have been near %@.", ptdesc]];
      
        NSRunAlertPanel(NSLocalizedString(@"route", @""), 
                              NSLocalizedString(@"About arrived!", @""), 
                                    NSLocalizedString(@"OK", @""), nil, nil); 
      }
    } 
  }
}

- (void)sayAlert:(NSString *)alert
{
  if (voiceAlerts) {
    NSPasteboard *pboard = [NSPasteboard pasteboardWithName: @"SpeechServicePasteboard"]; 

    [pboard declareTypes: [NSArray arrayWithObject: NSStringPboardType] owner: nil];
    [pboard setString: alert forType: NSStringPboardType];
    NSPerformService(@"Speech Output", pboard);
  }
}

- (void)setSayAlerts:(BOOL)value
{
  voiceAlerts = value;
}

@end



















